from functools import partial
from typing import Any, Callable, Dict, Optional, Tuple, Union

import chex
import jax
import jax.numpy as jnp
from gymnax.environments.environment import Environment, TEnvParams
from gymnax.environments.spaces import Box, Discrete
from jax2d.engine import PhysicsEngine
from jax2d.sim_state import CollisionManifold

from kinetix.environment.env_state import EnvParams, EnvState, StaticEnvParams
from kinetix.environment.spaces import (
    ActionType,
    ContinuousActions,
    DiscreteActions,
    EntityObservations,
    KinetixAction,
    KinetixObservation,
    MultiDiscreteActions,
    PixelObservations,
    SymbolicObservations,
    SymbolicPaddedObservations,
    ObservationType,
)

from kinetix.environment.utils import create_empty_env


class KinetixEnv(Environment):
    def __init__(
        self,
        action_type: KinetixAction,
        observation_type: KinetixObservation,
        static_env_params: StaticEnvParams,
        reset_function: Callable[[chex.PRNGKey], EnvState] = None,
        physics_engine: PhysicsEngine = None,
        auto_reset: bool = True,
    ):
        super().__init__()
        self.static_env_params = static_env_params
        self.action_type = action_type
        self.observation_type = observation_type
        self.physics_engine = physics_engine or PhysicsEngine(self.static_env_params)

        self.reset_function = reset_function
        self.auto_reset = auto_reset

    # Overridden functions from Gymnax
    def step_env(self, rng, state, action: jnp.ndarray, env_params):
        action_processed = self.action_type.process_action(action, state, self.static_env_params)
        return self.engine_step(state, action_processed, env_params)

    def reset_env(self, rng, env_params, override_reset_state: EnvState):
        rng, _rng_reset = jax.random.split(rng, 2)
        if override_reset_state is not None:
            env_state = override_reset_state
        elif self.reset_function is not None:
            env_state = self.reset_function(_rng_reset)
        else:
            raise NotImplementedError("No reset function provided")

        return self.get_obs(env_state), env_state

    # Copied from gymnax so we could add kwargs
    @partial(jax.jit, static_argnums=(0,))
    def step(
        self,
        key: chex.PRNGKey,
        state: EnvState,
        action: Union[int, float, chex.Array],
        env_params: Optional[TEnvParams] = None,
        override_reset_state: Optional[EnvState] = None,
    ) -> Tuple[chex.Array, EnvState, jnp.ndarray, jnp.ndarray, Dict[Any, Any]]:
        """Performs step transitions in the environment."""
        # Use default env parameters if no others specified
        if env_params is None:
            env_params = self.default_params
        key, key_reset = jax.random.split(key)

        obs_st, state_st, reward, done, info = self.step_env(key, state, action, env_params)
        if self.auto_reset:
            obs_re, state_re = self.reset_env(key_reset, env_params, override_reset_state=override_reset_state)
            # Auto-reset environment based on termination
            state = jax.tree.map(lambda x, y: jax.lax.select(done, x, y), state_re, state_st)
            obs = jax.tree.map(lambda x, y: jax.lax.select(done, x, y), obs_re, obs_st)
        else:
            obs = obs_st
            state = state_st

        return obs, state, reward, done, info

    @partial(jax.jit, static_argnums=(0,))
    def reset(
        self,
        key: chex.PRNGKey,
        env_params: Optional[TEnvParams] = None,
        override_reset_state: Optional[EnvState] = None,
    ) -> Tuple[chex.Array, EnvState]:
        """Performs resetting of environment."""
        # Use default env parameters if no others specified
        if env_params is None:
            env_params = self.default_params

        obs, state = self.reset_env(key, env_params, override_reset_state=override_reset_state)
        return obs, state

    # Actual Jax2D engine step
    @partial(jax.jit, static_argnums=(0,))
    def engine_step(self, env_state: EnvState, action_to_perform, env_params):
        def _single_step(env_state, unused):
            env_state, mfolds = self.physics_engine.step(
                env_state,
                env_params,
                action_to_perform,
            )

            reward, info = self.compute_reward_info(env_state, mfolds)

            done = reward != 0

            return env_state, (reward, done, info)

        env_state, (rewards, dones, infos) = jax.lax.scan(
            _single_step, env_state, xs=None, length=self.static_env_params.frame_skip
        )
        env_state = env_state.replace(timestep=env_state.timestep + 1)

        has_at_least_one_done = dones.sum() > 0
        first_done_index = dones.argmax()
        # If one of the steps led to a done, we should use the reward of the first one.
        # The rewards.max() has a problem if the first step has a reward of -1 and the second of 0.0
        reward = jax.lax.select(
            has_at_least_one_done,
            rewards[first_done_index],
            rewards.sum(),
        )
        done = has_at_least_one_done | jax.tree.reduce(
            jnp.logical_or, jax.tree.map(lambda x: jnp.isnan(x).any(), env_state), False
        )
        done |= env_state.timestep >= env_params.max_timesteps

        info = jax.tree.map(lambda x: jax.lax.select(has_at_least_one_done, x[first_done_index], x[-1]), infos)

        # dense reward
        delta_dist = (
            -(info["distance"] - env_state.last_distance) * env_params.dense_reward_scale
        )  # if distance got less, then reward is positive

        delta_dist = jnp.nan_to_num(delta_dist, nan=0.0, posinf=0.0, neginf=0.0)
        reward = reward + jax.lax.select(
            (env_state.last_distance == -1) | (env_params.dense_reward_scale == 0.0), 0.0, delta_dist
        )

        distance = jax.lax.select(done, -1.0, info["distance"])
        env_state = env_state.replace(last_distance=distance)

        return (
            jax.lax.stop_gradient(self.get_obs(env_state)),
            jax.lax.stop_gradient(env_state),
            reward,
            done,
            info,
        )

    # Information for dense rewards
    def compute_reward_info(
        self, state: EnvState, manifolds: tuple[CollisionManifold, CollisionManifold, CollisionManifold]
    ) -> float:
        def get_active(manifold: CollisionManifold) -> jnp.ndarray:
            return manifold.active

        def dist(a, b):
            return jnp.linalg.norm(a - b)

        @jax.vmap
        def dist_rr(idxa, idxb):
            return dist(state.polygon.position[idxa], state.polygon.position[idxb])

        @jax.vmap
        def dist_cc(idxa, idxb):
            return dist(state.circle.position[idxa], state.circle.position[idxb])

        @jax.vmap
        def dist_cr(idxa, idxb):
            return dist(state.circle.position[idxa], state.polygon.position[idxb])

        info = {
            "GoalR": False,
        }
        negative_reward = 0
        reward = 0
        distance = 0
        rs = state.polygon_shape_roles * state.polygon.active
        cs = state.circle_shape_roles * state.circle.active

        # Polygon Polygon
        r1 = rs[self.physics_engine.poly_poly_pairs[:, 0]]
        r2 = rs[self.physics_engine.poly_poly_pairs[:, 1]]
        reward += ((r1 * r2 == 2) * get_active(manifolds[0])).sum()
        negative_reward += ((r1 * r2 == 3) * get_active(manifolds[0])).sum()

        distance += (
            (r1 * r2 == 2)
            * dist_rr(self.physics_engine.poly_poly_pairs[:, 0], self.physics_engine.poly_poly_pairs[:, 1])
        ).sum()

        # Circle Polygon
        c1 = cs[self.physics_engine.circle_poly_pairs[:, 0]]
        r2 = rs[self.physics_engine.circle_poly_pairs[:, 1]]
        reward += ((c1 * r2 == 2) * get_active(manifolds[1])).sum()
        negative_reward += ((c1 * r2 == 3) * get_active(manifolds[1])).sum()

        t = dist_cr(self.physics_engine.circle_poly_pairs[:, 0], self.physics_engine.circle_poly_pairs[:, 1])
        distance += ((c1 * r2 == 2) * t).sum()

        # Circle Circle
        c1 = cs[self.physics_engine.circle_circle_pairs[:, 0]]
        c2 = cs[self.physics_engine.circle_circle_pairs[:, 1]]
        reward += ((c1 * c2 == 2) * get_active(manifolds[2])).sum()
        negative_reward += ((c1 * c2 == 3) * get_active(manifolds[2])).sum()

        distance += (
            (c1 * c2 == 2)
            * dist_cc(self.physics_engine.circle_circle_pairs[:, 0], self.physics_engine.circle_circle_pairs[:, 1])
        ).sum()

        reward = jax.lax.select(
            negative_reward > 0,
            -1.0,
            jax.lax.select(
                reward > 0,
                1.0,
                0.0,
            ),
        )

        info["GoalR"] = reward > 0
        info["distance"] = distance
        return reward, info

    # Action / Observations
    def action_space(self, env_params: Optional[EnvParams] = None) -> Union[Discrete, Box]:
        return self.action_type.action_space(env_params)

    def observation_space(self, env_params: EnvParams):
        return self.observation_type.observation_space(env_params)

    def get_obs(self, state: EnvState):
        return self.observation_type.get_obs(state)

    # Default env_params
    @property
    def default_params(self) -> EnvParams:
        return EnvParams()

    @staticmethod
    def default_static_params() -> StaticEnvParams:
        return StaticEnvParams()

    def __hash__(self):
        return hash(
            (
                hash(self.static_env_params),
                self.action_type.__class__.__name__,
                self.observation_type.__class__.__name__,
                self.reset_function,
                self.auto_reset,
            )
        )

    def __eq__(self, value):
        return hash(self) == hash(value)


class KinetixDummyEnv(KinetixEnv):
    def step_env(self, rng, state, action: jnp.ndarray, env_params):
        return (
            jax.lax.stop_gradient(self.get_obs(state)),
            jax.lax.stop_gradient(state),
            0.0,
            False,
            {"GoalR": False, "distance": 0.0},
        )

    def reset_env(self, rng, env_params, override_reset_state: EnvState):
        if override_reset_state is not None:
            env_state = override_reset_state
        elif self.reset_function is not None:
            env_state = create_empty_env(self.static_env_params)
        else:
            raise NotImplementedError("No reset function provided")

        return self.get_obs(env_state), env_state

    def __hash__(self):
        return hash(
            (
                hash(self.static_env_params),
                self.action_type.__class__.__name__,
                self.observation_type.__class__.__name__,
                self.reset_function,
                self.auto_reset,
                "dummy",
            )
        )

    def __eq__(self, value):
        return hash(self) == hash(value)


def make_kinetix_env(
    action_type: ActionType,
    observation_type: ObservationType,
    reset_fn: Optional[Callable[[chex.PRNGKey], EnvState]],
    env_params: Optional[EnvParams] = None,
    static_env_params: Optional[StaticEnvParams] = None,
    auto_reset: bool = True,
    create_dummy_env: bool = False,
) -> KinetixEnv:
    """

    Args:
        action_type (ActionType, optional): ActionType.CONTINUOUS or ActionType.MULTI_DISCRETE, overrides the option from the config if given. Defaults to None.
        observation_type (ObservationType, optional): ObservationType.PIXELS, ObservationType.SYMBOLIC_FLAT, ObservationType.SYMBOLIC_ENTITY, ObservationType.SYMBOLIC_FLAT_PADDED, overrides the option from the config if given. Defaults to None.
        reset_func (Callable[[chex.PRNGKey], EnvState], optional): If this is given, this is the function that gets called on reset to provide the starting state for the next episode. Defaults to None, in which case the environment has no auto reset behaviour
        env_params (EnvParams): EnvParams
        static_env_params (StaticEnvParams): StaticEnvParams
        auto_reset (bool): If True, the environment will automatically reset when the episode is done, in the standard gymnax way. Defaults to True.
        create_dummy_env (bool): If true, creates a dummy environment that has no-ops for the step and reset, so that it compiles significantly faster.

    Returns:
        KinetixEnv: The kinetix environment
    """

    if env_params is None:
        env_params = EnvParams()
    if static_env_params is None:
        static_env_params = StaticEnvParams()

    if action_type == ActionType.DISCRETE:
        action_type_cls = DiscreteActions
    elif action_type == ActionType.CONTINUOUS:
        action_type_cls = ContinuousActions
    elif action_type == ActionType.MULTI_DISCRETE:
        action_type_cls = MultiDiscreteActions
    else:
        raise ValueError(f"Invalid action_type '{action_type}', must be one of: [DISCRETE, CONTINUOUS, MULTI_DISCRETE]")
    action_type = action_type_cls(env_params, static_env_params)
    # observations
    if observation_type == ObservationType.PIXELS:
        obs_type_cls = PixelObservations
    elif observation_type == ObservationType.SYMBOLIC_FLAT:
        obs_type_cls = SymbolicObservations
    elif observation_type == ObservationType.SYMBOLIC_ENTITY:
        obs_type_cls = EntityObservations
    elif observation_type == ObservationType.SYMBOLIC_FLAT_PADDED:
        obs_type_cls = SymbolicPaddedObservations
    else:
        raise ValueError(
            f"Invalid observation_type '{observation_type}', must be one of: [pixels, symbolic_flat, symbolic_entity, symbolic_flat_padded]"
        )
    obs_type = obs_type_cls(env_params, static_env_params)

    env_cls = KinetixEnv
    if create_dummy_env:
        env_cls = KinetixDummyEnv
    return env_cls(
        action_type=action_type,
        observation_type=obs_type,
        static_env_params=static_env_params,
        reset_function=reset_fn,
        auto_reset=auto_reset,
    )



from functools import partial
from typing import Optional, Union

import chex
import jax
import jax.numpy as jnp
from jax2d.engine import calculate_collision_matrix, create_empty_sim
from jax2d.sim_state import SimState

from kinetix.environment.env_state import EnvParams, EnvState, StaticEnvParams


@partial(jax.jit, static_argnums=(0,))
def create_empty_env(static_env_params):
    sim_state = create_empty_sim(static_env_params)
    return EnvState(
        timestep=0,
        last_distance=-1.0,
        thruster_bindings=jnp.zeros(static_env_params.num_thrusters, dtype=jnp.int32),
        motor_bindings=jnp.zeros(static_env_params.num_joints, dtype=jnp.int32),
        motor_auto=jnp.zeros(static_env_params.num_joints, dtype=bool),
        polygon_shape_roles=jnp.zeros(static_env_params.num_polygons, dtype=jnp.int32),
        circle_shape_roles=jnp.zeros(static_env_params.num_circles, dtype=jnp.int32),
        polygon_highlighted=jnp.zeros(static_env_params.num_polygons, dtype=bool),
        circle_highlighted=jnp.zeros(static_env_params.num_circles, dtype=bool),
        polygon_densities=jnp.ones(static_env_params.num_polygons, dtype=jnp.float32),
        circle_densities=jnp.ones(static_env_params.num_circles, dtype=jnp.float32),
        **sim_state.__dict__,
    )


@jax.jit
def index_motor_actions(
    action: jnp.ndarray,
    state: EnvState,
    clip_min=None,
    clip_max=None,
):
    # Expand the motor actions to all joints with the same colour
    return jnp.clip(action[state.motor_bindings], clip_min, clip_max)


@jax.jit
def index_thruster_actions(
    action: jnp.ndarray,
    state: EnvState,
    clip_min=None,
    clip_max=None,
):
    # Expand the thruster actions to all joints with the same colour
    return jnp.clip(action[state.thruster_bindings], clip_min, clip_max)


@partial(jax.jit, static_argnums=(2,))
def convert_continuous_actions(
    action: jnp.ndarray, state: SimState, static_env_params: StaticEnvParams, env_params: EnvParams
):
    action_motor = action[: static_env_params.num_motor_bindings]
    action_thruster = action[static_env_params.num_motor_bindings :]
    action_motor = index_motor_actions(action_motor, state, -1, 1)
    action_thruster = index_thruster_actions(action_thruster, state, 0, 1)

    action_motor = jnp.where(state.motor_auto, jnp.ones_like(action_motor), action_motor)

    action_to_perform = jnp.concatenate([action_motor, action_thruster], axis=0)
    return action_to_perform


@partial(jax.jit, static_argnums=(2,))
def convert_discrete_actions(action: int, state: SimState, static_env_params: StaticEnvParams, env_params: EnvParams):
    # so, we have
    # 0 to NJC * 2 - 1: Joint Actions
    # NJC * 2: No-op
    # NJC * 2 + 1 to NJC * 2 + 1 + NTC - 1: Thruster Actions
    # action here is a categorical action
    which_idx = action // 2
    which_dir = action % 2
    actions = (
        jnp.zeros(static_env_params.num_motor_bindings + static_env_params.num_thruster_bindings)
        .at[which_idx]
        .set(which_dir * 2 - 1)
    )
    actions = actions * (
        1 - (action >= static_env_params.num_motor_bindings * 2)
    )  # if action is the last one, set it to zero, i.e., a no-op. Alternatively, if the action is larger than NJC * 2, then it is a thruster action and we shouldn't control the joints.

    actions = jax.lax.select(
        action > static_env_params.num_motor_bindings * 2,
        actions.at[action - static_env_params.num_motor_bindings * 2 - 1 + static_env_params.num_motor_bindings].set(1),
        actions,
    )

    action_motor = index_motor_actions(actions[: static_env_params.num_motor_bindings], state, -1, 1)
    action_motor = jnp.where(state.motor_auto, jnp.ones_like(action_motor), action_motor)
    action_thruster = index_thruster_actions(actions[static_env_params.num_motor_bindings :], state, 0, 1)
    action_to_perform = jnp.concatenate([action_motor, action_thruster], axis=0)
    return action_to_perform


@partial(jax.jit, static_argnums=(2,))
def convert_multi_discrete_actions(
    action: jnp.ndarray, state: SimState, static_env_params: StaticEnvParams, env_params: EnvParams
):
    # Comes in with each action being in {0,1,2} for joints and {0,1} for thrusters
    # Convert to [-1., 1.] for joints and [0., 1.] for thrusters

    def _single_motor_action(act):
        return jax.lax.switch(
            act,
            [lambda: 0.0, lambda: 1.0, lambda: -1.0],
        )

    def _single_thruster_act(act):
        return jax.lax.select(
            act == 0,
            0.0,
            1.0,
        )

    action_motor = jax.vmap(_single_motor_action)(action[: static_env_params.num_motor_bindings])
    action_thruster = jax.vmap(_single_thruster_act)(action[static_env_params.num_motor_bindings :])

    action_motor = index_motor_actions(action_motor, state, -1, 1)
    action_thruster = index_thruster_actions(action_thruster, state, 0, 1)

    action_motor = jnp.where(state.motor_auto, jnp.ones_like(action_motor), action_motor)

    action_to_perform = jnp.concatenate([action_motor, action_thruster], axis=0)
    return action_to_perform


@partial(jax.jit, static_argnums=(2,))
def permute_state(rng: chex.PRNGKey, env_state: EnvState, static_env_params: StaticEnvParams):
    idxs_circles = jnp.arange(static_env_params.num_circles)
    idxs_polygons = jnp.arange(static_env_params.num_polygons)
    idxs_joints = jnp.arange(static_env_params.num_joints)
    idxs_thrusters = jnp.arange(static_env_params.num_thrusters)

    rng, *_rngs = jax.random.split(rng, 5)
    idxs_circles_permuted = jax.random.permutation(_rngs[0], idxs_circles, independent=True)
    idxs_polygons_permuted = idxs_polygons.at[static_env_params.num_static_fixated_polys :].set(
        jax.random.permutation(_rngs[1], idxs_polygons[static_env_params.num_static_fixated_polys :], independent=True)
    )

    idxs_joints_permuted = jax.random.permutation(_rngs[2], idxs_joints, independent=True)
    idxs_thrusters_permuted = jax.random.permutation(_rngs[3], idxs_thrusters, independent=True)

    combined = jnp.concatenate([idxs_polygons_permuted, idxs_circles_permuted + static_env_params.num_polygons])
    # Change the ordering of the shapes, and also remember to change the indices associated with the joints

    inverse_permutation = jnp.argsort(combined)

    env_state = env_state.replace(
        polygon_shape_roles=env_state.polygon_shape_roles[idxs_polygons_permuted],
        circle_shape_roles=env_state.circle_shape_roles[idxs_circles_permuted],
        polygon_highlighted=env_state.polygon_highlighted[idxs_polygons_permuted],
        circle_highlighted=env_state.circle_highlighted[idxs_circles_permuted],
        polygon_densities=env_state.polygon_densities[idxs_polygons_permuted],
        circle_densities=env_state.circle_densities[idxs_circles_permuted],
        polygon=jax.tree.map(lambda x: x[idxs_polygons_permuted], env_state.polygon),
        circle=jax.tree.map(lambda x: x[idxs_circles_permuted], env_state.circle),
        joint=env_state.joint.replace(
            a_index=inverse_permutation[env_state.joint.a_index],
            b_index=inverse_permutation[env_state.joint.b_index],
        ),
        thruster=env_state.thruster.replace(
            object_index=inverse_permutation[env_state.thruster.object_index],
        ),
    )

    # And now permute the thrusters and joints
    env_state = env_state.replace(
        thruster_bindings=env_state.thruster_bindings[idxs_thrusters_permuted],
        motor_bindings=env_state.motor_bindings[idxs_joints_permuted],
        motor_auto=env_state.motor_auto[idxs_joints_permuted],
        joint=jax.tree.map(lambda x: x[idxs_joints_permuted], env_state.joint),
        thruster=jax.tree.map(lambda x: x[idxs_thrusters_permuted], env_state.thruster),
    )
    # and collision matrix
    env_state = env_state.replace(collision_matrix=calculate_collision_matrix(static_env_params, env_state.joint))
    return env_state



import jax
import jax.numpy as jnp
from jax2d.sim_state import RigidBody

from kinetix.environment.env_state import EnvParams, EnvState, StaticEnvParams


def _get_base_shape_features(
    density: jnp.ndarray, roles: jnp.ndarray, shapes: RigidBody, env_params: EnvParams
) -> jnp.ndarray:
    cos = jnp.cos(shapes.rotation)
    sin = jnp.sin(shapes.rotation)
    return jnp.concatenate(
        [
            shapes.position,
            shapes.velocity,
            jnp.expand_dims(shapes.inverse_mass, axis=1),
            jnp.expand_dims(shapes.inverse_inertia, axis=1),
            jnp.expand_dims(density, axis=1),
            jnp.expand_dims(jnp.tanh(shapes.angular_velocity / 10), axis=1),
            jax.nn.one_hot(roles, env_params.num_shape_roles),
            jnp.expand_dims(sin, axis=1),
            jnp.expand_dims(cos, axis=1),
            jnp.expand_dims(shapes.friction, axis=1),
            jnp.expand_dims(shapes.restitution, axis=1),
        ],
        axis=1,
    )


def add_circle_features(
    base_features: jnp.ndarray, shapes: RigidBody, env_params: EnvParams, static_env_params: StaticEnvParams
):
    return jnp.concatenate(
        [
            base_features,
            shapes.radius[:, None],
            jnp.ones_like(base_features[:, :1]),  # one for circle
        ],
        axis=1,
    )


def make_circle_features(
    state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams
) -> tuple[jnp.ndarray, jnp.ndarray]:
    base_features = _get_base_shape_features(state.circle_densities, state.circle_shape_roles, state.circle, env_params)
    node_features = add_circle_features(base_features, state.circle, env_params, static_env_params)
    return node_features, state.circle.active


def add_polygon_features(
    base_features: jnp.ndarray, shapes: RigidBody, env_params: EnvParams, static_env_params: StaticEnvParams
):
    vertices = jnp.where(
        jnp.arange(static_env_params.max_polygon_vertices)[None, :, None] < shapes.n_vertices[:, None, None],
        shapes.vertices,
        jnp.zeros_like(shapes.vertices) - 1,
    )

    return jnp.concatenate(
        [
            base_features,
            jnp.zeros_like(base_features[:, :1]),  # zero for polygon
            vertices.reshape((vertices.shape[0], -1)),
            jnp.expand_dims((shapes.n_vertices <= 3), axis=1),
        ],
        axis=1,
    )


def make_polygon_features(
    state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams
) -> tuple[jnp.ndarray, jnp.ndarray]:
    base_features = _get_base_shape_features(
        state.polygon_densities, state.polygon_shape_roles, state.polygon, env_params
    )
    node_features = add_polygon_features(base_features, state.polygon, env_params, static_env_params)
    return node_features, state.polygon.active


def make_unified_shape_features(
    state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams
) -> tuple[jnp.ndarray, jnp.ndarray]:
    base_p = _get_base_shape_features(state.polygon_densities, state.polygon_shape_roles, state.polygon, env_params)
    base_c = _get_base_shape_features(state.circle_densities, state.circle_shape_roles, state.circle, env_params)
    base_p = add_polygon_features(base_p, state.polygon, env_params, static_env_params)
    base_p = add_circle_features(base_p, state.polygon, env_params, static_env_params)

    base_c = add_polygon_features(base_c, state.circle, env_params, static_env_params)
    base_c = add_circle_features(base_c, state.circle, env_params, static_env_params)

    return jnp.concatenate([base_p, base_c], axis=0), jnp.concatenate(
        [state.polygon.active, state.circle.active], axis=0
    )


def make_joint_features(
    state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams
) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    # Returns joint_features, indexes, mask, of shape:
    # (2 * J, K), (2 * J, 2), (2 * J,)
    def _create_joint_features(joints):
        # 2, J, A
        J = joints.active.shape[0]

        def _create_1way_joint_features(direction):
            from_pos = jax.lax.select(direction, joints.a_relative_pos, joints.b_relative_pos)
            to_pos = jax.lax.select(direction, joints.b_relative_pos, joints.a_relative_pos)

            rotation_sin, rotation_cos = jnp.sin(joints.rotation), jnp.cos(joints.rotation)
            rotation_max_sin = jnp.sin(joints.max_rotation) * joints.motor_has_joint_limits
            rotation_max_cos = jnp.cos(joints.max_rotation) * joints.motor_has_joint_limits
            rotation_min_sin = jnp.sin(joints.min_rotation) * joints.motor_has_joint_limits
            rotation_min_cos = jnp.cos(joints.min_rotation) * joints.motor_has_joint_limits

            rotation_diff_max = (joints.max_rotation - joints.rotation) * joints.motor_has_joint_limits
            rotation_diff_min = (joints.min_rotation - joints.rotation) * joints.motor_has_joint_limits

            base_features = jnp.concatenate(
                [
                    (joints.active * 1.0)[:, None],
                    (joints.is_fixed_joint * 1.0)[:, None],  # J, 1
                    from_pos,
                    to_pos,
                    rotation_sin[:, None],
                    rotation_cos[:, None],
                ],
                axis=1,
            )
            rjoint_features = (
                jnp.concatenate(
                    [
                        joints.motor_speed[:, None],
                        joints.motor_power[:, None],
                        (joints.motor_on * 1.0)[:, None],
                        (joints.motor_has_joint_limits * 1.0)[:, None],
                        jax.nn.one_hot(state.motor_bindings, num_classes=static_env_params.num_motor_bindings),
                        rotation_min_sin[:, None],
                        rotation_min_cos[:, None],
                        rotation_max_sin[:, None],
                        rotation_max_cos[:, None],
                        rotation_diff_min[:, None],
                        rotation_diff_max[:, None],
                    ],
                    axis=1,
                )
                * (1.0 - (joints.is_fixed_joint * 1.0))[:, None]
            )

            return jnp.concatenate([base_features, rjoint_features], axis=1)

        # 2, J, A
        joint_features = jax.vmap(_create_1way_joint_features)(jnp.array([False, True]))

        # J, 2
        indexes_from = jnp.concatenate([joints.b_index[:, None], joints.a_index[:, None]], axis=1)
        indexes_to = jnp.concatenate([joints.a_index[:, None], joints.b_index[:, None]], axis=1)

        indexes_from = jnp.where(joints.active[:, None], indexes_from, jnp.zeros_like(indexes_from))
        indexes_to = jnp.where(joints.active[:, None], indexes_to, jnp.zeros_like(indexes_to))

        indexes = jnp.concatenate([indexes_from, indexes_to], axis=0)
        mask = jnp.concatenate([joints.active, joints.active], axis=0)

        return joint_features.reshape((2 * J, -1)), indexes, mask

    return _create_joint_features(state.joint)


def make_thruster_features(
    state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams
) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    # Returns thruster_features, indexes, mask, of shape:
    # (T, K), (T,), (T,)
    def _create_thruster_features(thrusters):
        cos = jnp.cos(thrusters.rotation)
        sin = jnp.sin(thrusters.rotation)
        return jnp.concatenate(
            [
                (thrusters.active * 1.0)[:, None],
                (thrusters.relative_position),
                jax.nn.one_hot(state.thruster_bindings, num_classes=static_env_params.num_thruster_bindings),
                sin[:, None],
                cos[:, None],
                thrusters.power[:, None],
            ],
            axis=1,
        )

    return _create_thruster_features(state.thruster), state.thruster.object_index, state.thruster.active



from gymnax import EnvParams
import jax
import jax.numpy as jnp
from jax2d.engine import select_shape
from jax2d.maths import rmat
from jax2d.sim_state import RigidBody, Thruster, Joint

from kinetix.environment.env_state import EnvState, StaticEnvParams


def features_to_rigidbody(
    env_params: EnvParams, static_env_params: StaticEnvParams, features: jnp.ndarray, nshapes: int, is_circle: bool
) -> tuple[RigidBody, jnp.ndarray, jnp.ndarray]:
    """Given the rigidbody features from a symbolic observation, convert it to a RigidBody object."""
    nr = env_params.num_shape_roles

    # Circle + Poly

    # shapes.position, 0:2
    # shapes.velocity, 2:4
    # jnp.expand_dims(shapes.inverse_mass, axis=1), 4
    # jnp.expand_dims(shapes.inverse_inertia, axis=1), 5
    # jnp.expand_dims(density, axis=1), 6
    # jnp.expand_dims(jnp.tanh(shapes.angular_velocity / 10), axis=1), 7
    # jax.nn.one_hot(roles, env_params.num_shape_roles), 8:8+roles
    # jnp.expand_dims(sin, axis=1), 8+roles
    # jnp.expand_dims(cos, axis=1), 9+roles
    # jnp.expand_dims(shapes.friction, axis=1), 10+roles
    # jnp.expand_dims(shapes.restitution, axis=1), 11+roles

    # Circle:
    # radius: 12+roles

    # Poly:
    # dummy one hot poly vs circle: 12 + roles
    # vertices: 13+roles : 13+roles + 2 * max_polygon_vertices
    # n_vertices: 13+roles + 2 * max_polygon_vertices

    sin = features[:, 8 + nr]
    cos = features[:, 9 + nr]
    rotation = jnp.atan2(sin, cos)
    rotation = jnp.where(rotation < 0, rotation + 2 * jnp.pi, rotation)

    rb_shape_roles = jnp.argmax(features[:, 8 : 8 + nr], axis=1)
    rb_densities = features[:, 6]

    rb = RigidBody(
        position=features[:, :2],
        velocity=features[:, 2:4],
        inverse_mass=features[:, 4],
        inverse_inertia=features[:, 5],
        angular_velocity=jnp.atanh(features[:, 7]) * 10.0,
        rotation=rotation,
        friction=features[:, 10 + nr],
        restitution=features[:, 11 + nr],
        collision_mode=jnp.ones(nshapes, dtype=jnp.int32),
        active=jnp.zeros(nshapes, dtype=jnp.bool),
        n_vertices=jnp.zeros(nshapes, dtype=jnp.int32),
        vertices=jnp.zeros((nshapes, static_env_params.max_polygon_vertices, 2)),
        radius=jnp.zeros(nshapes),
    )

    if is_circle:
        rb = rb.replace(
            radius=features[:, 12 + nr],
            active=features[:, 12 + nr] > 0.0,
        )
    else:

        mpv = static_env_params.max_polygon_vertices
        has_three_vertices = features[:, 13 + nr + 2 * mpv]

        rb = rb.replace(
            vertices=features[:, 13 + nr : 13 + nr + 2 * mpv].reshape((-1, mpv, 2)),
            n_vertices=jnp.where(
                has_three_vertices,
                jnp.ones_like(has_three_vertices, dtype=jnp.int32) * 3,
                jnp.ones_like(has_three_vertices, dtype=jnp.int32) * 4,
            ),
            collision_mode=rb.collision_mode.at[..., 0].set(2),
        )
        rb = rb.replace(active=jnp.abs(rb.vertices.reshape(rb.vertices.shape[0], -1)).sum(axis=1) > 0.0)
    return rb, rb_shape_roles, rb_densities


def features_to_thruster(
    static_env_params: StaticEnvParams,
    env_state: EnvState,
    thruster_features: jnp.ndarray,
    nshapes: int,
    thruster_idxs_override=None,
):
    # (thrusters.active * 1.0)[:, None], 0
    # (thrusters.relative_position), 1:3
    # jax.nn.one_hot(state.thruster_bindings, num_classes=static_env_params.num_thruster_bindings), 3:3+ntb
    # sin[:, None], 3+ntb
    # cos[:, None], 4+ntb
    # thrusters.power[:, None], 5+ntb
    # jax.nn.one_hot(thruster_idxs, nshapes) 5+ntb:5+ntb+nshapes

    ntb = static_env_params.num_thruster_bindings
    if thruster_idxs_override is None:
        thruster_idxs = jnp.argmax(thruster_features[:, 6 + ntb : 6 + ntb + nshapes], axis=1)
    else:
        thruster_idxs = thruster_idxs_override

    sin = thruster_features[:, 3 + ntb]
    cos = thruster_features[:, 4 + ntb]
    rotation = jnp.atan2(sin, cos)
    rotation = jnp.where(rotation < 0, rotation + 2 * jnp.pi, rotation)

    thruster_bindings = jnp.argmax(thruster_features[:, 3 : 3 + ntb], axis=1)

    thruster_shapes = jax.vmap(select_shape, in_axes=(None, 0, None))(env_state, thruster_idxs, static_env_params)

    active = thruster_features[:, 0] > 0.0

    thruster_relative_position = thruster_features[:, 1:3]
    thruster_global_position = thruster_shapes.position + jax.vmap(jnp.matmul)(
        jax.vmap(rmat)(thruster_shapes.rotation), thruster_relative_position
    )

    thruster_global_position = jnp.where(
        active[:, None], thruster_global_position, jnp.zeros_like(thruster_global_position)
    )
    return (
        Thruster(
            active=active,
            relative_position=thruster_relative_position,
            rotation=rotation,
            power=thruster_features[:, 5 + ntb],
            global_position=thruster_global_position,
            object_index=thruster_idxs,
        ),
        thruster_bindings,
    )


def features_to_joint(
    static_env_params: StaticEnvParams,
    env_state: EnvState,
    joint_features: jnp.ndarray,
    nshapes: int,
    joint_idxs_override=None,
):
    # (joints.active * 1.0)[:, None] 0
    # (joints.is_fixed_joint * 1.0)[:, None], 1
    # from_pos 2:4
    # to_pos 4:6
    # rotation_sin[:, None] 6
    # rotation_cos[:, None] 7
    # joints.motor_speed[:, None], 8
    # joints.motor_power[:, None], 9
    # (joints.motor_on * 1.0)[:, None], 10
    # (joints.motor_has_joint_limits * 1.0)[:, None], 11
    # jax.nn.one_hot(state.motor_bindings, num_classes=static_env_params.num_motor_bindings), 12:12+nmb
    # rotation_min_sin[:, None], 12+nmb
    # rotation_min_cos[:, None], 13+nmb
    # rotation_max_sin[:, None], 14+nmb
    # rotation_max_cos[:, None], 15+nmb
    # rotation_diff_min[:, None], 16+nmb
    # rotation_diff_max[:, None], 17+nmb
    # jax.nn.one_hot(joint_idxs[:J, 0], nshapes),  18+nmb:18+nmb+nshapes
    # jax.nn.one_hot(joint_idxs[:J, 1], nshapes),  18+nmb+nshapes:18+nmb+2*nshapes
    nmb = static_env_params.num_motor_bindings

    if joint_idxs_override is None:
        b_index = jnp.argmax(joint_features[:, 18 + nmb : 18 + nmb + nshapes], axis=1)
        a_index = jnp.argmax(joint_features[:, 18 + nmb + nshapes : 18 + nmb + nshapes * 2], axis=1)
    else:
        b_index = joint_idxs_override[:, 0]
        a_index = joint_idxs_override[:, 1]

    b_relative_pos = joint_features[:, 2:4]
    a_relative_pos = joint_features[:, 4:6]
    joint_rotation = jnp.arctan2(joint_features[:, 6], joint_features[:, 7])
    joint_rotation = jnp.where(joint_rotation < 0, joint_rotation + 2 * jnp.pi, joint_rotation)

    joint_shapes = jax.vmap(select_shape, in_axes=(None, 0, None))(env_state, a_index, static_env_params)
    active = joint_features[:, 0] > 0.0

    joint_global_position = joint_shapes.position + jax.vmap(jnp.matmul)(
        jax.vmap(rmat)(joint_shapes.rotation), a_relative_pos
    )
    joint_global_position = jnp.where(active[:, None], joint_global_position, jnp.zeros_like(joint_global_position))

    return (
        Joint(
            active=active,
            is_fixed_joint=joint_features[:, 1] > 0.0,
            a_index=a_index,
            b_index=b_index,
            a_relative_pos=a_relative_pos,
            b_relative_pos=b_relative_pos,
            rotation=joint_rotation,
            motor_speed=joint_features[:, 8],
            motor_power=joint_features[:, 9],
            motor_on=joint_features[:, 10] > 0.0,
            motor_has_joint_limits=joint_features[:, 11] > 0.0,
            min_rotation=jnp.arctan2(joint_features[:, 12 + nmb], joint_features[:, 13 + nmb]),
            max_rotation=jnp.arctan2(joint_features[:, 14 + nmb], joint_features[:, 15 + nmb]),
            global_position=joint_global_position,
            acc_r_impulse=jnp.zeros((static_env_params.num_joints), dtype=jnp.float32),
            acc_impulse=jnp.zeros((static_env_params.num_joints, 2), dtype=jnp.float32),
        ),
        jnp.argmax(joint_features[:, 12 : 12 + nmb], axis=1),
    )



import math
from functools import partial

import jax
import jax.numpy as jnp
from jax2d.engine import (
    calc_inverse_inertia_circle,
    calc_inverse_inertia_polygon,
    calc_inverse_mass_circle,
    calc_inverse_mass_polygon,
    calculate_collision_matrix,
    recalculate_mass_and_inertia,
    select_shape,
)
from jax2d.maths import rmat
from jax2d.sim_state import Joint, RigidBody, Thruster

from kinetix.environment.env_state import EnvParams, EnvState, StaticEnvParams
from kinetix.environment.ued.ued_state import UEDParams
from kinetix.environment.ued.util import (
    are_there_shapes_present,
    count_roles,
    get_role,
    is_space_for_joint,
    is_space_for_shape,
    make_do_dummy_step,
    random_position_on_circle,
    random_position_on_polygon,
    sample_dimensions,
)


@partial(jax.jit, static_argnums=(3, 4))
def mutate_add_shape(
    rng,
    state: EnvState,
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    ued_params: UEDParams,
    force_no_fixate: bool = False,
):
    def do_dummy(rng, state):
        return state

    def do_add(rng, state):
        rng, _rng = jax.random.split(rng)
        _rngs = jax.random.split(_rng, 9)

        space_for_new_rect = state.polygon.active.astype(int).sum() < static_env_params.num_polygons
        space_for_new_circle = state.circle.active.astype(int).sum() < static_env_params.num_circles

        is_rect_p = jnp.array([space_for_new_rect * 1.0, space_for_new_circle * 1.0])
        is_rect = jax.random.choice(_rngs[0], jnp.array([True, False], dtype=bool), p=is_rect_p)

        rect_index = jnp.argmin(state.polygon.active)
        circle_index = jnp.argmin(state.circle.active)

        shape_role = get_role(_rngs[1], state, static_env_params)

        max_shape_size = (
            jnp.array([1.0, ued_params.goal_body_size_factor, ued_params.goal_body_size_factor, 1.0])[shape_role]
            * ued_params.max_shape_size
        )

        vertices, half_dimensions, radius = sample_dimensions(
            _rngs[2],
            static_env_params,
            is_rect,
            ued_params,
            max_shape_size=max_shape_size,
        )
        n_vertices = jax.lax.select(ued_params.generate_triangles, jax.random.choice(_rngs[3], jnp.array([3, 4])), 4)

        largest = jnp.max(jnp.array([half_dimensions[0] * jnp.sqrt(2), half_dimensions[1] * jnp.sqrt(2), radius]))

        screen_dim_world = (
            static_env_params.screen_dim[0] / env_params.pixels_per_unit,
            static_env_params.screen_dim[1] / env_params.pixels_per_unit,
        )
        min_x = largest
        max_x = screen_dim_world[0] - largest
        min_y = largest + 0.4
        max_y = screen_dim_world[1] - largest

        def _og_minmax():
            return min_x, max_x, min_y, max_y

        def _opposite_minmax():
            return jax.lax.switch(
                shape_role,
                [
                    (lambda: (min_x, max_x, min_y, max_y)),
                    (lambda: (min_x, max_x - screen_dim_world[0] / 2, min_y, max_y)),
                    (lambda: (min_x + screen_dim_world[0] / 2, max_x, min_y, max_y)),
                    (lambda: (min_x, max_x, min_y, max_y)),
                ],
            )

        min_x, max_x, min_y, max_y = jax.lax.cond(
            jax.random.uniform(_rngs[4], shape=()) < ued_params.goal_body_opposide_side_chance,
            _opposite_minmax,
            _og_minmax,
        )

        position = jax.random.uniform(_rngs[5], shape=(2,)) * jnp.array(
            [
                max_x - min_x,
                max_y - min_y,
            ]
        ) + jnp.array([min_x, min_y])

        rotation = jax.random.uniform(_rngs[6], shape=()) * 2 * math.pi
        velocity = jnp.array([0.0, 0.0])
        angular_velocity = 0.0

        density = 1.0
        inverse_mass = jax.lax.select(
            is_rect,
            calc_inverse_mass_polygon(vertices, n_vertices, static_env_params, density)[0],
            calc_inverse_mass_circle(radius, density),
        )

        inverse_inertia = jax.lax.select(
            is_rect,
            calc_inverse_inertia_polygon(vertices, n_vertices, static_env_params, density),
            calc_inverse_inertia_circle(radius, density),
        )

        fixate_chance = ued_params.fixate_chance_min + (1.0 / inverse_mass) * ued_params.fixate_chance_scale
        fixate_chance = jnp.minimum(fixate_chance, ued_params.fixate_chance_max)
        is_fixated = jax.random.uniform(_rngs[7], shape=()) < fixate_chance
        is_fixated &= ~force_no_fixate

        inverse_mass *= 1 - is_fixated
        inverse_inertia *= 1 - is_fixated

        # We want to bias fixated shapes to starting nearer the bottom half of the screen
        fixate_shape_bottom_bias = (
            ued_params.fixate_shape_bottom_bias + ued_params.fixate_shape_bottom_bias_special_role * (shape_role != 0)
        )
        is_forcing_bottom = jax.random.uniform(_rngs[8]) < fixate_shape_bottom_bias

        half_screen_height = (static_env_params.screen_dim[1] / env_params.pixels_per_unit) / 2.0
        position = jax.lax.select(
            is_fixated & is_forcing_bottom & (position[1] >= half_screen_height),
            position.at[1].add(-half_screen_height),
            position,
        )

        # This could be either a rect or a circle
        new_rigid_body = RigidBody(
            position=position,
            velocity=velocity,
            inverse_mass=inverse_mass,
            inverse_inertia=inverse_inertia,
            rotation=rotation,
            angular_velocity=angular_velocity,
            radius=radius,
            active=True,
            friction=1.0,
            vertices=vertices,
            n_vertices=n_vertices,
            collision_mode=1,
            restitution=0.0,
        )

        state = state.replace(
            polygon=jax.tree.map(
                lambda x, y: jax.lax.select(is_rect, y.at[rect_index].set(x), y), new_rigid_body, state.polygon
            ),
            circle=jax.tree.map(
                lambda x, y: jax.lax.select(jnp.logical_not(is_rect), y.at[circle_index].set(x), y),
                new_rigid_body,
                state.circle,
            ),
            polygon_shape_roles=jax.lax.select(
                is_rect,
                state.polygon_shape_roles.at[rect_index].set(shape_role),
                state.polygon_shape_roles,
            ),
            circle_shape_roles=jax.lax.select(
                jnp.logical_not(is_rect),
                state.circle_shape_roles.at[circle_index].set(shape_role),
                state.circle_shape_roles,
            ),
        )
        return recalculate_mass_and_inertia(state, static_env_params, state.polygon_densities, state.circle_densities)

    return jax.lax.cond(is_space_for_shape(state), do_add, do_dummy, rng, state)


@partial(jax.jit, static_argnums=(3, 4))
def mutate_add_connected_shape(
    rng,
    state: EnvState,
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    ued_params: UEDParams,
    force_rjoint: bool = False,
):
    def do_dummy(rng, state):
        return state, False

    def do_add(rng, state):
        rng, _rng = jax.random.split(rng)
        _rngs = jax.random.split(_rng, 21)

        # Select a random index amongst the currently active shapes.
        p_rect = state.polygon.active.at[: static_env_params.num_static_fixated_polys].set(False)
        p_circle = state.circle.active

        p_rect = p_rect.astype(jnp.float32)
        p_circle = p_circle.astype(jnp.float32)

        p_rect *= (state.polygon.inverse_mass == 0) * ued_params.connect_to_fixated_prob_coeff + (
            state.polygon.inverse_mass != 0
        ) * 1.0
        p_circle *= (state.circle.inverse_mass == 0) * ued_params.connect_to_fixated_prob_coeff + (
            state.circle.inverse_mass != 0
        ) * 1.0

        # Bias based on number of existing connections
        rect_connections = jnp.zeros(static_env_params.num_polygons)
        circle_connections = jnp.zeros(static_env_params.num_circles)

        rect_connections = rect_connections.at[state.joint.a_index].add(
            jnp.ones(static_env_params.num_joints)
            * state.joint.active
            * (state.joint.a_index < static_env_params.num_polygons)
        )
        rect_connections = rect_connections.at[state.joint.b_index].add(
            jnp.ones(static_env_params.num_joints)
            * state.joint.active
            * (state.joint.b_index < static_env_params.num_polygons)
        )

        circle_connections = circle_connections.at[state.joint.a_index - static_env_params.num_polygons].add(
            jnp.ones(static_env_params.num_joints)
            * state.joint.active
            * (state.joint.a_index >= static_env_params.num_polygons)
        )
        circle_connections = circle_connections.at[state.joint.b_index - static_env_params.num_polygons].add(
            jnp.ones(static_env_params.num_joints)
            * state.joint.active
            * (state.joint.b_index >= static_env_params.num_polygons)
        )

        # Rectangles can have up to 2 connections
        p_rect *= (-rect_connections + 2.0) / 2.0
        p_rect = jnp.maximum(p_rect, 0.0)
        # Circles can have 1 connection
        p_circle *= circle_connections == 0

        # To sample a target rect/circle, we have to have at least one.
        target_rect_p = jnp.array(
            [
                (state.polygon.active.astype(int).sum() > static_env_params.num_static_fixated_polys) * 1.0,
                (state.circle.active.astype(int).sum() > 0) * 1.0,
            ]
        )

        # Don't connect to a circle if no connection-free ones exist
        target_rect_p = target_rect_p.at[1].mul(p_circle.sum() > 0)

        space_for_new_rect = state.polygon.active.astype(int).sum() < static_env_params.num_polygons
        space_for_new_circle = state.circle.active.astype(int).sum() < static_env_params.num_circles

        is_target_rect = jax.random.choice(_rngs[0], jnp.array([True, False], dtype=bool), p=target_rect_p) | (
            ~space_for_new_rect
        )

        is_rect_p = jnp.array([space_for_new_rect * 1.0, space_for_new_circle * 1.0])
        is_rect = jax.random.choice(_rngs[1], jnp.array([True, False], dtype=bool), p=is_rect_p) | (
            ~is_target_rect & space_for_new_rect
        )

        shape_index = jax.lax.select(
            is_rect,
            jnp.argmin(state.polygon.active),
            jnp.argmin(state.circle.active),
        )
        unified_shape_index = shape_index + (~is_rect) * static_env_params.num_polygons

        vertices, half_dimensions, radius = sample_dimensions(
            _rngs[2], static_env_params, is_rect, ued_params, max_shape_size=ued_params.max_shape_size
        )
        n_vertices = jax.lax.select(ued_params.generate_triangles, jax.random.choice(_rngs[3], jnp.array([3, 4])), 4)

        rotation = jax.random.uniform(_rngs[4], shape=()) * 2 * math.pi
        velocity = jnp.array([0.0, 0.0])
        angular_velocity = 0.0

        density = 1.0
        inverse_mass = jax.lax.select(
            is_rect,
            calc_inverse_mass_polygon(vertices, n_vertices, static_env_params, density)[0],
            calc_inverse_mass_circle(radius, density),
        )

        inverse_inertia = jax.lax.select(
            is_rect,
            calc_inverse_inertia_polygon(vertices, n_vertices, static_env_params, density),
            calc_inverse_inertia_circle(radius, density),
        )

        # Joint

        current_num_rjoints = (jnp.logical_not(state.joint.is_fixed_joint) * state.joint.active).sum()
        is_rjoint = jnp.logical_or(
            jnp.logical_or(jax.random.uniform(_rngs[5]) < 0.5, force_rjoint),
            current_num_rjoints < ued_params.min_rjoints_bias,
        )

        joint_index = jnp.argmin(state.joint.active)

        local_joint_position_rect = random_position_on_polygon(_rngs[6], vertices, n_vertices, static_env_params)
        local_joint_position_circle = random_position_on_circle(_rngs[7], radius, on_centre_chance=1.0)

        local_joint_position = jax.lax.select(is_rect, local_joint_position_rect, local_joint_position_circle)

        p_rect = jax.lax.select(p_rect.sum() == 0, state.polygon.active.astype(jnp.float32), p_rect)
        p_circle = jax.lax.select(p_circle.sum() == 0, state.circle.active.astype(jnp.float32), p_circle)

        target_index = jax.lax.select(
            is_target_rect,
            jax.random.choice(
                _rngs[8],
                jnp.arange(static_env_params.num_polygons),
                p=p_rect,
            ),
            jax.random.choice(
                _rngs[9],
                jnp.arange(static_env_params.num_circles),
                p=p_circle,
            ),
        )

        unified_target_index = target_index + jnp.logical_not(is_target_rect) * static_env_params.num_polygons
        target_shape = select_shape(state, unified_target_index, static_env_params)

        target_joint_position_rect = random_position_on_polygon(
            _rngs[10], state.polygon.vertices[target_index], state.polygon.n_vertices[target_index], static_env_params
        )
        target_joint_position_circle = random_position_on_circle(
            _rngs[11], state.circle.radius[target_index], on_centre_chance=1.0
        )

        target_joint_position = jax.lax.select(is_target_rect, target_joint_position_rect, target_joint_position_circle)

        # Calculate the world position of the new shape
        # We know the rotation of the new shape. We also know the position of the current shape, which we want to remain fixed.
        # Set `position` such that local_joint_position is the same as `target_joint_position`
        global_joint_pos = target_shape.position + jnp.matmul(rmat(target_shape.rotation), target_joint_position)
        position = global_joint_pos - jnp.matmul(rmat(rotation), local_joint_position)

        _, pos_diff = calc_inverse_mass_polygon(vertices, n_vertices, static_env_params, density)
        position = jax.lax.select(is_rect, position + pos_diff, position)
        local_joint_position = jax.lax.select(is_rect, local_joint_position - pos_diff, local_joint_position)
        vertices = jax.lax.select(is_rect, vertices - pos_diff[None], vertices)

        target_role = jax.lax.select(
            is_target_rect, state.polygon_shape_roles[target_index], state.circle_shape_roles[target_index]
        )

        # We cannot have role 1 and role 2 being connected.
        p = jnp.array([1.0, 1.0, 1.0, 1.0])
        # If role is 0, keep all probs at 1, otherwise set the target role's complement to 0 prob
        # 3 - role turns 1 to 2 and 2 to 1
        # If the target role is three, we set everything to zero except for the default
        p = jax.lax.select(
            target_role == 0,
            p,
            jax.lax.select(
                target_role <= 2,
                p.at[3 - target_role].set(False).at[3].set(False),
                (p.at[2].set(False).at[1].set(False)),
            ),
        )

        shape_role = get_role(_rngs[12], state, static_env_params, initial_p=p)

        # This could be either a rect or a circle
        new_rigid_body = RigidBody(
            position=position,
            velocity=velocity,
            inverse_mass=inverse_mass,
            inverse_inertia=inverse_inertia,
            rotation=rotation,
            angular_velocity=angular_velocity,
            radius=radius,
            active=True,
            friction=1.0,
            vertices=vertices,
            n_vertices=n_vertices,
            collision_mode=1,
            restitution=0.0,
        )

        # Change the shape indices such that a_index is less than b_index
        a_index = shape_index + (1 - is_rect) * static_env_params.num_polygons
        b_index = target_index + (1 - is_target_rect) * static_env_params.num_polygons

        should_swap = a_index > b_index
        a_index, b_index, local_joint_position, target_joint_position, shape_a, shape_b = jax.lax.cond(
            should_swap,
            lambda x: (x[1], x[0], x[3], x[2], x[5], x[4]),  # pairwise swap
            lambda x: x,
            (a_index, b_index, local_joint_position, target_joint_position, new_rigid_body, target_shape),
        )

        motor_on = jax.random.uniform(_rngs[13], shape=()) < ued_params.motor_on_chance
        joint_colour = jax.random.randint(_rngs[14], shape=(), minval=0, maxval=static_env_params.num_motor_bindings)
        joint_rotation = shape_b.rotation - shape_a.rotation

        motor_speed = jax.random.uniform(
            _rngs[15], shape=(), minval=ued_params.motor_min_speed, maxval=ued_params.motor_max_speed
        )

        motor_power = jax.random.uniform(
            _rngs[16], shape=(), minval=ued_params.motor_min_power, maxval=ued_params.motor_max_power
        )
        wheel_power = jax.random.uniform(
            _rngs[20], shape=(), minval=ued_params.motor_min_power, maxval=ued_params.wheel_max_power
        )

        # High-powered wheels break the physics engine - this is a temporary fix
        motor_power = jax.lax.select(is_rect & is_target_rect, motor_power, wheel_power)

        motor_has_joint_limits = jax.random.uniform(_rngs[17], shape=()) < ued_params.joint_limit_chance
        motor_has_joint_limits &= is_rect & is_target_rect
        joint_limit_min = (
            jax.random.uniform(_rngs[18], shape=(), minval=-ued_params.joint_limit_max, maxval=0.0)
            * motor_has_joint_limits
        )
        joint_limit_max = (
            jax.random.uniform(_rngs[19], shape=(), minval=0.0, maxval=ued_params.joint_limit_max)
            * motor_has_joint_limits
        )

        rjoint = Joint(
            a_index=a_index,
            b_index=b_index,
            a_relative_pos=local_joint_position,
            b_relative_pos=target_joint_position,
            global_position=global_joint_pos,
            active=True,
            motor_speed=motor_speed,
            motor_power=motor_power,
            motor_on=motor_on,
            # colour=joint_colour,
            motor_has_joint_limits=motor_has_joint_limits,
            min_rotation=joint_limit_min,
            max_rotation=joint_limit_max,
            is_fixed_joint=False,
            rotation=0.0,
            acc_impulse=jnp.zeros((2,), dtype=jnp.float32),
            acc_r_impulse=jnp.zeros((), dtype=jnp.float32),
        )

        fjoint = Joint(
            a_index=a_index,
            b_index=b_index,
            a_relative_pos=local_joint_position,
            b_relative_pos=target_joint_position,
            global_position=global_joint_pos,
            active=True,
            rotation=joint_rotation,
            acc_impulse=jnp.zeros((2,), dtype=jnp.float32),
            acc_r_impulse=jnp.zeros((), dtype=jnp.float32),
            is_fixed_joint=True,
            motor_has_joint_limits=False,
            min_rotation=0.0,
            max_rotation=0.0,
            motor_on=False,
            motor_power=0.0,
            motor_speed=0.0,
        )

        state = state.replace(
            polygon=jax.tree.map(
                lambda x, y: jax.lax.select(is_rect, y.at[shape_index].set(x), y), new_rigid_body, state.polygon
            ),
            circle=jax.tree.map(
                lambda x, y: jax.lax.select(jnp.logical_not(is_rect), y.at[shape_index].set(x), y),
                new_rigid_body,
                state.circle,
            ),
            joint=jax.tree.map(
                lambda rj, fj, y: jax.lax.select(is_rjoint, y.at[joint_index].set(rj), y.at[joint_index].set(fj)),
                rjoint,
                fjoint,
                state.joint,
            ),
            polygon_shape_roles=jax.lax.select(
                is_rect,
                state.polygon_shape_roles.at[shape_index].set(shape_role),
                state.polygon_shape_roles,
            ),
            circle_shape_roles=jax.lax.select(
                jnp.logical_not(is_rect),
                state.circle_shape_roles.at[shape_index].set(shape_role),
                state.circle_shape_roles,
            ),
            motor_bindings=state.motor_bindings.at[joint_index].set(joint_colour),
        )

        # We need the new collision matrix.
        state = state.replace(collision_matrix=calculate_collision_matrix(static_env_params, state.joint))
        state = recalculate_mass_and_inertia(state, static_env_params, state.polygon_densities, state.circle_densities)

        # Was this a valid addition?
        # We calculate whether (assuming the possiblity of 360 degree rotation around the joint)
        # both shapes can be visible
        # This is to remove the common degenerate pattern of connected shapes being fully inside each other
        def _get_min_rect_dist(r_id, local_pos):
            rect: RigidBody = jax.tree.map(lambda x: x[r_id], state.polygon)

            half_width = (jnp.max(rect.vertices[:, 0]) - jnp.min(rect.vertices[:, 0])) / 2.0
            half_height = (jnp.max(rect.vertices[:, 1]) - jnp.min(rect.vertices[:, 1])) / 2.0

            dist_x = half_width - jnp.abs(local_pos[0])
            dist_y = half_height - jnp.abs(local_pos[1])

            return jnp.minimum(dist_x, dist_y)

        def _get_max_rect_dist(r_id, local_pos):
            rect: RigidBody = jax.tree.map(lambda x: x[r_id], state.polygon)

            half_width = (jnp.max(rect.vertices[:, 0]) - jnp.min(rect.vertices[:, 0])) / 2.0
            half_height = (jnp.max(rect.vertices[:, 1]) - jnp.min(rect.vertices[:, 1])) / 2.0

            dist_x = jnp.maximum(
                jnp.abs(half_width - local_pos[0]),
                jnp.abs(-half_width - local_pos[0]),
            )

            dist_y = jnp.maximum(
                jnp.abs(half_height - local_pos[1]),
                jnp.abs(-half_height - local_pos[1]),
            )

            return jnp.sqrt(dist_x * dist_x + dist_y * dist_y)

        def are_both_shapes_showing(idx1, idx2, local_pos1, local_pos2):
            def _is_small_shape_showing(small_idx, big_idx, small_local_pos, big_local_pos):
                small_is_poly = small_idx < static_env_params.num_polygons
                big_is_poly = big_idx < static_env_params.num_polygons

                # CC
                cc_result = False

                # CR
                cr_r_dist = _get_min_rect_dist(big_idx, big_local_pos)
                cr_result = (
                    cr_r_dist + ued_params.connect_visibility_min
                    < state.circle.radius[small_idx - static_env_params.num_polygons]
                )

                # RC
                rc_r_dist = _get_max_rect_dist(small_idx, small_local_pos)
                rc_result = (
                    rc_r_dist
                    > state.circle.radius[big_idx - static_env_params.num_polygons] + ued_params.connect_visibility_min
                )

                # RR
                rr_small_dist = _get_max_rect_dist(small_idx, small_local_pos)
                rr_big_dist = _get_min_rect_dist(big_idx, big_local_pos)
                rr_result = rr_small_dist > rr_big_dist + ued_params.connect_visibility_min

                # Select
                return jax.lax.select(
                    small_is_poly,
                    jax.lax.select(big_is_poly, rr_result, rc_result),
                    jax.lax.select(big_is_poly, cr_result, cc_result),
                )

            # Are both shapes showing?
            return _is_small_shape_showing(idx1, idx2, local_pos1, local_pos2) & _is_small_shape_showing(
                idx2, idx1, local_pos2, local_pos1
            )

        valid = are_both_shapes_showing(
            unified_shape_index, unified_target_index, local_joint_position, target_joint_position
        )
        return state, valid

    # To add a connected shape, we must have both at least one existing shape and space
    return jax.lax.cond(
        is_space_for_shape(state) & are_there_shapes_present(state, static_env_params) & is_space_for_joint(state),
        do_add,
        do_dummy,
        rng,
        state,
    )


@partial(jax.jit, static_argnums=(3, 4))
def mutate_add_connected_shape_proper(
    rng,
    state: EnvState,
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    ued_params: UEDParams,
    force_rjoint: bool = False,
):
    return mutate_add_connected_shape(rng, state, env_params, static_env_params, ued_params, force_rjoint=force_rjoint)[
        0
    ]


@partial(jax.jit, static_argnums=(3, 4))
def mutate_remove_shape(
    rng, state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams, ued_params: UEDParams
):

    can_remove_mask = (
        jnp.concatenate([state.polygon.active, state.circle.active])
        .at[: static_env_params.num_static_fixated_polys]
        .set(False)
    )

    def dummy(rng, state):
        return state

    def do_remove(rng, state: EnvState):
        rng, _rng = jax.random.split(rng)
        rngs = jax.random.split(_rng, 2)
        p = can_remove_mask.astype(jnp.float32)
        index_to_remove = jax.random.choice(rngs[0], jnp.arange(can_remove_mask.shape[0]), p=p)
        is_rect = index_to_remove < static_env_params.num_polygons
        state = state.replace(
            polygon=state.polygon.replace(
                active=jax.lax.select(
                    is_rect, state.polygon.active.at[index_to_remove].set(False), state.polygon.active
                )
            ),
            circle=state.circle.replace(
                active=jax.lax.select(
                    jnp.logical_not(is_rect),
                    state.circle.active.at[index_to_remove - static_env_params.num_polygons].set(False),
                    state.circle.active,
                )
            ),
        )
        # We need to now remove any joints connected to this shape
        joints_to_remove = (state.joint.a_index == index_to_remove) | (state.joint.b_index == index_to_remove)

        thrusters_to_remove = state.thruster.object_index == index_to_remove

        state = state.replace(
            joint=state.joint.replace(active=jnp.where(joints_to_remove, False, state.joint.active)),
            thruster=state.thruster.replace(active=jnp.where(thrusters_to_remove, False, state.thruster.active)),
        )
        # Now recalculate collision matrix
        state = state.replace(collision_matrix=calculate_collision_matrix(static_env_params, state.joint))
        return state

    return jax.lax.cond(can_remove_mask.sum() > 0, do_remove, dummy, rng, state)


@partial(jax.jit, static_argnums=(3, 4))
def mutate_remove_joint(
    rng, state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams, ued_params: UEDParams
):
    can_remove_mask = state.joint.active

    def dummy(rng, state):
        return state

    def do_remove(rng, state):
        rng, _rng = jax.random.split(rng)
        rngs = jax.random.split(_rng, 2)
        p = can_remove_mask.astype(jnp.float32)
        index_to_remove = jax.random.choice(rngs[0], jnp.arange(can_remove_mask.shape[0]), p=p)
        state = state.replace(joint=state.joint.replace(active=state.joint.active.at[index_to_remove].set(False)))
        # Recalculate collision matrix.
        state = state.replace(collision_matrix=calculate_collision_matrix(static_env_params, state.joint))
        return state

    return jax.lax.cond(can_remove_mask.sum() > 0, do_remove, dummy, rng, state)


@partial(jax.jit, static_argnums=(3, 4))
def mutate_swap_role(
    rng, state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams, ued_params: UEDParams
):
    def _cr(*args):
        return count_roles(*args, include_static_polys=False)

    role_counts = jax.vmap(_cr, (None, None, 0))(state, static_env_params, jnp.arange(4))
    are_there_multiple_roles = (role_counts > 0).sum() > 1

    def dummy(rng, state):
        return state

    def do_swap(rng, state):
        rng, _rng = jax.random.split(rng)
        rngs = jax.random.split(_rng, 2)
        all_roles = jnp.concatenate([state.polygon_shape_roles, state.circle_shape_roles])

        p = (
            (jnp.concatenate([state.polygon.active, state.circle.active]))
            .astype(jnp.float32)
            .at[: static_env_params.num_static_fixated_polys]
            .set(0.0)
        )
        shape_idx_a = jax.random.choice(
            rngs[0], jnp.arange(static_env_params.num_polygons + static_env_params.num_circles), p=p
        )
        role_a = all_roles[shape_idx_a]
        p = jnp.where(all_roles == role_a, 0.0, p)
        shape_idx_b = jax.random.choice(
            rngs[1], jnp.arange(static_env_params.num_polygons + static_env_params.num_circles), p=p
        )
        role_b = all_roles[shape_idx_b]
        role_a, role_b = role_b, role_a

        for idx, role in [(shape_idx_a, role_a), (shape_idx_b, role_b)]:
            is_rect = idx < static_env_params.num_polygons
            state = state.replace(
                polygon_shape_roles=jax.lax.select(
                    is_rect, state.polygon_shape_roles.at[idx].set(role), state.polygon_shape_roles
                ),
                circle_shape_roles=jax.lax.select(
                    jnp.logical_not(is_rect),
                    state.circle_shape_roles.at[idx - static_env_params.num_polygons].set(role),
                    state.circle_shape_roles,
                ),
            )
        return state

    return jax.lax.cond(are_there_multiple_roles, do_swap, dummy, rng, state)


@partial(jax.jit, static_argnums=(3, 4))
def mutate_toggle_fixture(
    rng, state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams, ued_params: UEDParams
):
    can_toggle_mask = (
        jnp.concatenate([state.polygon.active, state.circle.active])
        .at[: static_env_params.num_static_fixated_polys]
        .set(False)
    )

    def dummy(rng, state):
        return state

    def do_toggle(rng, state: EnvState):
        rng, _rng = jax.random.split(rng)
        rngs = jax.random.split(_rng, 2)
        p = can_toggle_mask.astype(jnp.float32)
        index_to_remove = jax.random.choice(rngs[0], jnp.arange(can_toggle_mask.shape[0]), p=p)
        is_rect = index_to_remove < static_env_params.num_polygons
        is_current_fixed = (
            jax.lax.select(
                is_rect,
                state.polygon.inverse_inertia[index_to_remove],
                state.circle.inverse_inertia[index_to_remove - static_env_params.num_polygons],
            )
            == 0.0
        )

        is_current_fixed = is_current_fixed * 1.0  # if it is fixed, we set it to 1.0 and recalc.
        # If it is not fixed, this is 0.0, and it makes it fixed.

        state = state.replace(
            polygon=state.polygon.replace(
                inverse_inertia=jax.lax.select(
                    is_rect,
                    state.polygon.inverse_inertia.at[index_to_remove].set(is_current_fixed),
                    state.polygon.inverse_inertia,
                ),
                inverse_mass=jax.lax.select(
                    is_rect,
                    state.polygon.inverse_mass.at[index_to_remove].set(is_current_fixed),
                    state.polygon.inverse_mass,
                ),
            ),
            circle=state.circle.replace(
                inverse_inertia=jax.lax.select(
                    jnp.logical_not(is_rect),
                    state.circle.inverse_inertia.at[index_to_remove - static_env_params.num_polygons].set(
                        is_current_fixed
                    ),
                    state.circle.inverse_inertia,
                ),
                inverse_mass=jax.lax.select(
                    jnp.logical_not(is_rect),
                    state.circle.inverse_mass.at[index_to_remove - static_env_params.num_polygons].set(
                        is_current_fixed
                    ),
                    state.circle.inverse_mass,
                ),
            ),
        )

        state = recalculate_mass_and_inertia(state, static_env_params, state.polygon_densities, state.circle_densities)
        return state

    return jax.lax.cond(can_toggle_mask.sum() > 0, do_toggle, dummy, rng, state)


@partial(jax.jit, static_argnums=(3, 4))
def mutate_add_thruster(
    rng, state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams, ued_params: UEDParams
):
    is_fixated = jnp.concatenate([state.polygon.inverse_mass == 0, state.circle.inverse_mass == 0])
    # is_fixated = jnp.zeros_like(is_fixated, dtype=bool)
    is_active = jnp.concatenate([state.polygon.active, state.circle.active])
    can_add_mask = is_active & (~is_fixated)
    can_add_mask = jnp.logical_and(is_active, jnp.logical_not(is_fixated))

    def dummy(rng, state):
        return state

    def do_add(rng, state: EnvState):
        rng, _rng = jax.random.split(rng)
        _rngs = jax.random.split(_rng, 10)
        p = can_add_mask.astype(jnp.float32)
        shape_index = jax.random.choice(_rngs[0], jnp.arange(can_add_mask.shape[0]), p=p)
        thruster_idx = jnp.argmin(state.thruster.active)

        shape = select_shape(state, shape_index, static_env_params)

        position_to_add_thruster = jax.lax.select(
            shape_index < static_env_params.num_polygons,
            random_position_on_polygon(_rngs[1], shape.vertices, shape.n_vertices, static_env_params),
            random_position_on_circle(_rngs[2], shape.radius, on_centre_chance=0.0),
        )

        direction_to_com = ((jax.random.uniform(_rngs[3]) > 0.5) * 2 - 1) * position_to_add_thruster
        direction_to_com = jax.lax.select(
            jnp.linalg.norm(direction_to_com) == 0.0, jnp.array([1.0, 0.0]), direction_to_com
        )

        thruster_angle = jax.lax.select(
            jax.random.uniform(_rngs[4]) < ued_params.thruster_align_com_prob,
            jnp.atan2(direction_to_com[1], direction_to_com[0]),  # test this
            jax.random.uniform(
                _rngs[5],
                (),
            )
            * 2
            * jnp.pi,
        )

        thruster_power = jax.random.uniform(_rngs[6]) * 1.5 + 0.5

        thruster = Thruster(
            object_index=shape_index,
            active=True,
            relative_position=position_to_add_thruster,  # jnp.array([0.0, 0.0]),  # a bit of a hack but reasonable.
            rotation=thruster_angle,  # jax.random.choice(rngs[1], jnp.arange(4) * jnp.pi / 2),
            power=1.0
            / jax.lax.select(shape.inverse_mass == 0, 1.0, shape.inverse_mass)
            * ued_params.thruster_power_multiplier
            * thruster_power,
            global_position=shape.position + jnp.matmul(rmat(shape.rotation), position_to_add_thruster),
        )
        thruster_colour = jax.random.randint(
            _rngs[7], shape=(), minval=0, maxval=static_env_params.num_thruster_bindings
        )

        state = state.replace(
            thruster=jax.tree.map(lambda y, x: y.at[thruster_idx].set(x), state.thruster, thruster),
            thruster_bindings=state.thruster_bindings.at[thruster_idx].set(thruster_colour),
        )

        return state

    return jax.lax.cond(
        jnp.logical_and((can_add_mask.sum() > 0), (jnp.logical_not(state.thruster.active).sum() > 0)),
        do_add,
        dummy,
        rng,
        state,
    )


@partial(jax.jit, static_argnums=(3, 4))
def mutate_change_gravity(
    rng, state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams, ued_params: UEDParams
):
    rng, _rng = jax.random.split(rng)
    rngs = jax.random.split(_rng, 2)
    new_gravity = jax.lax.select(
        jax.random.uniform(rngs[0]) < 0.5,
        jnp.array([0.0, -9.8]),
        jnp.array([0.0, jax.random.uniform(rngs[1], minval=-9.8, maxval=0)]),
    )

    return state.replace(gravity=new_gravity)


@partial(jax.jit, static_argnums=(3, 4))
def mutate_remove_thruster(
    rng, state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams, ued_params: UEDParams
):
    are_there_thrusters = state.thruster.active

    def dummy(rng, state):
        return state

    def do_remove(rng, state):
        rng, _rng = jax.random.split(rng)
        rngs = jax.random.split(_rng, 2)
        p = are_there_thrusters.astype(jnp.float32)
        thruster_idx = jax.random.choice(rngs[0], jnp.arange(are_there_thrusters.shape[0]), p=p)
        return state.replace(thruster=state.thruster.replace(active=state.thruster.active.at[thruster_idx].set(False)))

    return jax.lax.cond(are_there_thrusters.sum() > 0, do_remove, dummy, rng, state)


def make_mutate_change_shape_size(env_params, static_env_params):
    do_dummy_step = make_do_dummy_step(env_params, static_env_params)

    @partial(jax.jit, static_argnums=(3, 4))
    def mutate_change_shape_size(
        rng, state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams, ued_params: UEDParams
    ):
        shape_active = jnp.concatenate(
            [state.polygon.active.at[: static_env_params.num_static_fixated_polys].set(False), state.circle.active]
        )

        def dummy(rng, state):
            return state

        def do_change(rng, state):
            rng, _rng = jax.random.split(rng)
            rngs = jax.random.split(_rng, 10)
            p = shape_active.astype(jnp.float32)
            shape_idx = jax.random.choice(rngs[0], jnp.arange(shape_active.shape[0]), p=p)
            is_rect = shape_idx < static_env_params.num_polygons
            vertices, _, radius = sample_dimensions(
                rngs[1], static_env_params, is_rect, ued_params, max_shape_size=ued_params.max_shape_size
            )

            idx_new_top_left = jnp.argmin(vertices[:, 0] * 100 + vertices[:, 1])
            idx_old_top_left = jnp.argmin(
                state.polygon.vertices[shape_idx, :, 0] * 100 + state.polygon.vertices[shape_idx, :, 1]
            )
            scale_rect = (vertices[idx_new_top_left]) / (state.polygon.vertices[shape_idx, idx_old_top_left])
            scale_circle = radius / state.circle.radius[shape_idx - static_env_params.num_polygons]
            vertices = state.polygon.vertices[shape_idx] * scale_rect

            scale = jax.lax.select(
                is_rect,
                scale_rect,
                jnp.array([scale_circle, scale_circle]),
            )

            is_a = ((state.joint.a_index == shape_idx) & state.joint.active)[:, None]
            is_b = ((state.joint.b_index == shape_idx) & state.joint.active)[:, None]
            state = state.replace(
                joint=state.joint.replace(
                    a_relative_pos=(state.joint.a_relative_pos * scale[None]) * is_a
                    + (1 - is_a) * state.joint.a_relative_pos,
                    b_relative_pos=(state.joint.b_relative_pos * scale[None]) * is_b
                    + (1 - is_b) * state.joint.b_relative_pos,
                ),
                polygon=state.polygon.replace(
                    vertices=jax.lax.select(
                        is_rect, state.polygon.vertices.at[shape_idx].set(vertices), state.polygon.vertices
                    ),
                ),
                circle=state.circle.replace(
                    radius=jax.lax.select(
                        jnp.logical_not(is_rect),
                        state.circle.radius.at[shape_idx - static_env_params.num_polygons].set(radius),
                        state.circle.radius,
                    )
                ),
            )

            def _ss(state, _):
                return do_dummy_step(state), None

            state = jax.lax.scan(_ss, state, jnp.arange(5))[0]
            return recalculate_mass_and_inertia(
                state, static_env_params, state.polygon_densities, state.circle_densities
            )

        return jax.lax.cond(shape_active.sum() > 0, do_change, dummy, rng, state)

    return mutate_change_shape_size


@partial(jax.jit, static_argnums=(3, 4))
def mutate_change_shape_location(
    rng, state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams, ued_params: UEDParams
):
    shape_active = jnp.concatenate(
        [state.polygon.active.at[: static_env_params.num_static_fixated_polys].set(False), state.circle.active]
    )

    def dummy(rng, state):
        return state

    def do_change(rng, state):
        rng, _rng = jax.random.split(rng)
        rngs = jax.random.split(_rng, 10)
        p = shape_active.astype(jnp.float32)
        shape_idx = jax.random.choice(rngs[0], jnp.arange(shape_active.shape[0]), p=p)
        delta_pos = jax.random.uniform(rngs[1], shape=(2,)) - 0.5  # [-0.5, 0.5]

        positions = jnp.concatenate([state.polygon.position, state.circle.position])

        mask_of_shape_locations_to_change = (
            (state.collision_matrix[shape_idx] == 0).at[: static_env_params.num_static_fixated_polys].set(False)
        )
        # check the new positions, but then maybe revert if any shape becomes out of bounds now.
        new_positions_tentative = positions * (
            1 - mask_of_shape_locations_to_change[:, None]
        ) + mask_of_shape_locations_to_change[:, None] * (positions + delta_pos[None])

        polys = state.polygon
        p_pos = new_positions_tentative[: static_env_params.num_polygons]
        c_pos = new_positions_tentative[static_env_params.num_polygons :]  # state.circle.position
        rad = state.circle.radius
        rect_vertex_mask = jnp.arange(static_env_params.max_polygon_vertices)[None] < polys.n_vertices[:, None]
        rect_mask = polys.active.at[: static_env_params.num_static_fixated_polys].set(False)
        circ_mask = state.circle.active
        # check if new pos maybe goes out of bounds:
        min_x, max_x, min_y, max_y = (
            jnp.minimum(
                jnp.min(
                    p_pos[:, 0] + jnp.min(polys.vertices[:, :, 0], where=rect_vertex_mask, initial=0, axis=1),
                    where=rect_mask,
                    initial=jnp.inf,
                ),
                jnp.min(c_pos[:, 0] - rad, where=circ_mask, initial=jnp.inf),
            ),
            jnp.maximum(
                jnp.max(
                    p_pos[:, 0] + jnp.max(polys.vertices[:, :, 0], where=rect_vertex_mask, initial=0, axis=1),
                    where=rect_mask,
                    initial=-jnp.inf,
                ),
                jnp.max(c_pos[:, 0] + rad, where=circ_mask, initial=-jnp.inf),
            ),
            jnp.minimum(
                jnp.min(
                    p_pos[:, 1] + jnp.min(polys.vertices[:, :, 1], where=rect_vertex_mask, initial=0, axis=1),
                    where=rect_mask,
                    initial=jnp.inf,
                ),
                jnp.min(c_pos[:, 1] - rad, where=circ_mask, initial=jnp.inf),
            ),
            jnp.maximum(
                jnp.max(
                    p_pos[:, 1] + jnp.max(polys.vertices[:, :, 1], where=rect_vertex_mask, initial=0, axis=1),
                    where=rect_mask,
                    initial=-jnp.inf,
                ),
                jnp.max(c_pos[:, 1] + rad, where=circ_mask, initial=-jnp.inf),
            ),
        )

        how_much_oob_x_left = jnp.maximum(0, 0 - min_x)
        how_much_oob_x_right = jnp.maximum(0, max_x - static_env_params.screen_dim[0] / env_params.pixels_per_unit)
        how_much_oob_y_down = jnp.maximum(0, 0.4 - min_y)  # this is for the floor
        how_much_oob_y_up = jnp.maximum(0, max_y - static_env_params.screen_dim[1] / env_params.pixels_per_unit)

        # correct by out of bounds factor
        positions = (
            new_positions_tentative
            + jnp.array(
                [
                    how_much_oob_x_left - how_much_oob_x_right,
                    how_much_oob_y_down - how_much_oob_y_up,
                ]
            )[None]
            * mask_of_shape_locations_to_change[:, None]
        )

        state = state.replace(
            polygon=state.polygon.replace(
                position=positions[: static_env_params.num_polygons],
            ),
            circle=state.circle.replace(
                position=positions[static_env_params.num_polygons :],
            ),
        )
        return recalculate_mass_and_inertia(state, static_env_params, state.polygon_densities, state.circle_densities)

    return jax.lax.cond(shape_active.sum() > 0, do_change, dummy, rng, state)


def make_mutate_change_shape_rotation(env_params, static_env_params):
    do_dummy_step = make_do_dummy_step(env_params, static_env_params)

    @partial(jax.jit, static_argnums=(3, 4))
    def mutate_change_shape_rotation(
        rng, state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams, ued_params: UEDParams
    ):
        shape_active = jnp.concatenate(
            [state.polygon.active.at[: static_env_params.num_static_fixated_polys].set(False), state.circle.active]
        )

        def dummy(rng, state):
            return state

        def do_change(rng, state):
            rng, _rng = jax.random.split(rng)
            rngs = jax.random.split(_rng, 10)
            p = shape_active.astype(jnp.float32)
            shape_idx = jax.random.choice(rngs[0], jnp.arange(shape_active.shape[0]), p=p)
            is_rect = shape_idx < static_env_params.num_polygons

            rotation_delta = jax.random.uniform(rngs[1], shape=()) * math.pi / 2

            has_fixed_joint_a = (state.joint.a_index == shape_idx) & state.joint.is_fixed_joint & state.joint.active
            has_fixed_joint_b = (state.joint.b_index == shape_idx) & state.joint.is_fixed_joint & state.joint.active

            state = state.replace(
                joint=state.joint.replace(
                    rotation=jax.lax.select(
                        has_fixed_joint_a,
                        state.joint.rotation - rotation_delta,
                        jax.lax.select(
                            has_fixed_joint_b,
                            state.joint.rotation + rotation_delta,
                            state.joint.rotation,
                        ),
                    )
                ),
                polygon=state.polygon.replace(
                    rotation=jax.lax.select(
                        is_rect, state.polygon.rotation.at[shape_idx].add(rotation_delta), state.polygon.rotation
                    ),
                ),
                circle=state.circle.replace(
                    rotation=jax.lax.select(
                        jnp.logical_not(is_rect),
                        state.circle.rotation.at[shape_idx - static_env_params.num_polygons].add(rotation_delta),
                        state.circle.rotation,
                    )
                ),
            )

            def _ss(state, _):
                return do_dummy_step(state), None

            state = jax.lax.scan(_ss, state, jnp.arange(5))[0]
            return recalculate_mass_and_inertia(
                state, static_env_params, state.polygon_densities, state.circle_densities
            )

        return jax.lax.cond(shape_active.sum() > 0, do_change, dummy, rng, state)

    return mutate_change_shape_rotation



import dataclasses
import math

from flax import struct


@struct.dataclass
class UEDParams:
    max_shape_size: float = 1.0
    goal_body_opposide_side_chance: float = 0.5
    goal_body_size_factor: float = 1.0
    min_rjoints_bias: int = 2

    large_rect_dim_chance: float = 0.3
    large_rect_dim_scale: float = 2.0

    generate_triangles: bool = False
    thruster_power_multiplier: float = 2.0

    thruster_align_com_prob: float = 0.8

    motor_on_chance: float = 0.8
    motor_min_speed: float = 0.4
    motor_max_speed: float = 3.0
    motor_min_power: float = 1.0
    motor_max_power: float = 3.0
    wheel_max_power: float = 1.0

    joint_limit_chance: float = 0.4
    joint_limit_max: float = math.pi
    joint_fixed_chance: float = 0.1

    fixate_chance_min: float = 0.02
    fixate_chance_max: float = 1.0
    fixate_chance_scale: float = 4.0  # Fixation probability scales with size
    fixate_shape_bottom_bias: float = 0.0
    fixate_shape_bottom_bias_special_role: float = 0.6

    circle_max_size_coeff: float = 0.8

    connect_to_fixated_prob_coeff: float = 0.05
    connect_visibility_min: float = 0.05
    connect_no_visibility_bias: float = 10.0

    add_shape_chance: float = 0.35
    add_connected_shape_chance: float = 0.35
    add_no_shape_chance: float = 0.3
    add_thruster_chance: float = 0.3
    add_shape_n_proposals: int = 8

    floor_prob_normal: float = 0.9
    floor_prob_green: float = 0.0
    floor_prob_blue: float = 0.02
    floor_prob_red: float = 0.08

    def __hash__(self):
        field_values = {f.name: getattr(self, f.name) for f in dataclasses.fields(self)}
        return hash(tuple(sorted(field_values.items())))

    def __eq__(self, value):
        return super().__eq__(value)



# Copyright 2023 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# CODE IS HEAVILY INSPIRED FROM https://github.com/huggingface/transformers/blob/v4.40.1/src/transformers/models/deprecated/transfo_xl/modeling_transfo_xl.py
# MOST OF THE TIME JUST A CONVERSION IN JAX


"""Relative Attention HEAVILY INSPIRED FROM https://github.com/huggingface/transformers/blob/v4.40.1/src/transformers/models/deprecated/transfo_xl/modeling_transfo_xl.py
, flax attention, https://github.com/kimiyoung/transformer-xl/blob/master/pytorch/mem_transformer.py#L143, most of the time just a flax/jax conversion """

import functools
from typing import Any, Callable, Optional, Tuple

import jax
import jax.numpy as jnp
from flax.linen import initializers
from flax.linen.dtypes import promote_dtype
from flax.linen.linear import DenseGeneral, DotGeneralT, PrecisionLike, default_kernel_init
from flax.linen.module import Module, compact, merge_param
from jax import lax, random

PRNGKey = Any
Shape = Tuple[int, ...]
Dtype = Any
Array = Any

roll_vmap = jax.vmap(jnp.roll, in_axes=(-2, 0, None), out_axes=-2)


def _rel_shift(x):
    zero_pad_shape = x.shape[:-2] + (x.shape[-2], 1)
    zero_pad = jnp.zeros(zero_pad_shape, dtype=x.dtype)
    x_padded = jnp.concatenate([zero_pad, x], axis=-1)

    x_padded_shape = x.shape[:-2] + (x.shape[-1] + 1, x.shape[-2])
    x_padded = x_padded.reshape(x_padded_shape)
    # x_padded=jnp.swapaxes(x_padded,0,1)

    x = jnp.take(x_padded, jnp.arange(1, x_padded.shape[-2]), axis=-2).reshape(x.shape)

    return x


def dot_product_attention_weights(
    query: Array,
    key: Array,
    r_pos_embed,
    r_r_bias,
    r_w_bias,
    bias: Optional[Array] = None,
    mask: Optional[Array] = None,
    broadcast_dropout: bool = True,
    dropout_rng: Optional[PRNGKey] = None,
    dropout_rate: float = 0.0,
    deterministic: bool = False,
    dtype: Optional[Dtype] = None,
    precision: PrecisionLike = None,
):
    """Computes dot-product attention weights given query and key.

    Used by :func:`dot_product_attention`, which is what you'll most likely use.
    But if you want access to the attention weights for introspection, then
    you can directly call this function and call einsum yourself.

    Args:
        query: queries for calculating attention with shape of
            `[batch..., q_length, num_heads, qk_depth_per_head]`.
        key: keys for calculating attention with shape of
            `[batch..., kv_length, num_heads, qk_depth_per_head]`.
        bias: bias for the attention weights. This should be broadcastable to the
            shape `[batch..., num_heads, q_length, kv_length]`.
            This can be used for incorporating causal masks, padding masks,
            proximity bias, etc.
        mask: mask for the attention weights. This should be broadcastable to the
            shape `[batch..., num_heads, q_length, kv_length]`.
            This can be used for incorporating causal masks.
            Attention weights are masked out if their corresponding mask value
            is `False`.
        broadcast_dropout: bool: use a broadcasted dropout along batch dims.
        dropout_rng: JAX PRNGKey: to be used for dropout
        dropout_rate: dropout rate
        deterministic: bool, deterministic or not (to apply dropout)
        dtype: the dtype of the computation (default: infer from inputs and params)
        precision: numerical precision of the computation see `jax.lax.Precision`
            for details.

    Returns:
        Output of shape `[batch..., num_heads, q_length, kv_length]`.
    """
    query, key = promote_dtype(query, key, dtype=dtype)
    dtype = query.dtype

    assert query.ndim == key.ndim, "q, k must have same rank."
    assert query.shape[:-3] == key.shape[:-3], "q, k batch dims must match."
    assert query.shape[-2] == key.shape[-2], "q, k num_heads must match."
    assert query.shape[-1] == key.shape[-1], "q, k depths must match."

    # calculate attention matrix
    depth = query.shape[-1]
    # query = query

    # attn weight shape is (batch..., num_heads, q_length, kv_length)
    attn_weights = jnp.einsum("...qhd,...khd->...hqk", query + r_w_bias, key, precision=precision)

    attn_weights_r = jnp.einsum("...qhd,khd->...hqk", query + r_r_bias, r_pos_embed, precision=precision)

    attn_weights_r = roll_vmap(attn_weights_r, jnp.arange(0, query.shape[-3]) - (query.shape[-3] - 1), -1)
    # attn_weights_r=_rel_shift(attn_weights_r)
    attn_weights = attn_weights + attn_weights_r

    attn_weights = attn_weights / jnp.sqrt(depth).astype(dtype)

    # apply attention bias: masking, dropout, proximity bias, etc.
    if bias is not None:
        attn_weights = attn_weights + bias
    # apply attention mask
    if mask is not None:
        big_neg = jnp.finfo(dtype).min
        attn_weights = jnp.where(mask, attn_weights, big_neg)

    # normalize the attention weights
    attn_weights = jax.nn.softmax(attn_weights).astype(dtype)

    # apply attention dropout
    if not deterministic and dropout_rate > 0.0:
        keep_prob = 1.0 - dropout_rate
        if broadcast_dropout:
            # dropout is broadcast across the batch + head dimensions
            dropout_shape = tuple([1] * (key.ndim - 2)) + attn_weights.shape[-2:]
            keep = random.bernoulli(dropout_rng, keep_prob, dropout_shape)  # type: ignore
        else:
            keep = random.bernoulli(dropout_rng, keep_prob, attn_weights.shape)  # type: ignore
        multiplier = keep.astype(dtype) / jnp.asarray(keep_prob, dtype=dtype)
        attn_weights = attn_weights * multiplier

    return attn_weights


def dot_product_attention(
    query: Array,
    key: Array,
    value: Array,
    r_pos_embed,
    r_r_bias,
    r_w_bias,
    bias: Optional[Array] = None,
    mask: Optional[Array] = None,
    broadcast_dropout: bool = True,
    dropout_rng: Optional[PRNGKey] = None,
    dropout_rate: float = 0.0,
    deterministic: bool = False,
    dtype: Optional[Dtype] = None,
    precision: PrecisionLike = None,
):
    """Computes dot-product attention given query, key, and value.

    This is the core function for applying attention based on
    https://arxiv.org/abs/1706.03762. It calculates the attention weights given
    query and key and combines the values using the attention weights.

    Note: query, key, value needn't have any batch dimensions.

    Args:
        query: queries for calculating attention with shape of
            `[batch..., q_length, num_heads, qk_depth_per_head]`.
        key: keys for calculating attention with shape of
            `[batch..., kv_length, num_heads, qk_depth_per_head]`.
        value: values to be used in attention with shape of
            `[batch..., kv_length, num_heads, v_depth_per_head]`.
        bias: bias for the attention weights. This should be broadcastable to the
            shape `[batch..., num_heads, q_length, kv_length]`.
            This can be used for incorporating causal masks, padding masks,
            proximity bias, etc.
        mask: mask for the attention weights. This should be broadcastable to the
            shape `[batch..., num_heads, q_length, kv_length]`.
            This can be used for incorporating causal masks.
            Attention weights are masked out if their corresponding mask value
            is `False`.
        broadcast_dropout: bool: use a broadcasted dropout along batch dims.
        dropout_rng: JAX PRNGKey: to be used for dropout
        dropout_rate: dropout rate
        deterministic: bool, deterministic or not (to apply dropout)
        dtype: the dtype of the computation (default: infer from inputs)
        precision: numerical precision of the computation see `jax.lax.Precision`
            for details.

    Returns:
        Output of shape `[batch..., q_length, num_heads, v_depth_per_head]`.
    """
    query, key, value = promote_dtype(query, key, value, dtype=dtype)
    dtype = query.dtype
    assert key.ndim == query.ndim == value.ndim, "q, k, v must have same rank."
    assert query.shape[:-3] == key.shape[:-3] == value.shape[:-3], "q, k, v batch dims must match."
    assert query.shape[-2] == key.shape[-2] == value.shape[-2], "q, k, v num_heads must match."
    assert key.shape[-3] == value.shape[-3], "k, v lengths must match."

    # compute attention weights
    attn_weights = dot_product_attention_weights(
        query,
        key,
        r_pos_embed,
        r_r_bias,
        r_w_bias,
        bias,
        mask,
        broadcast_dropout,
        dropout_rng,
        dropout_rate,
        deterministic,
        dtype,
        precision,
    )

    # return weighted sum over values for each query position
    return jnp.einsum("...hqk,...khd->...qhd", attn_weights, value, precision=precision)


class RelMultiHeadDotProductAttention(Module):
    """Multi-head dot-product attention.

    Attributes:
        num_heads: number of attention heads. Features (i.e. inputs_q.shape[-1])
            should be divisible by the number of heads.
        dtype: the dtype of the computation (default: infer from inputs and params)
        param_dtype: the dtype passed to parameter initializers (default: float32)
        qkv_features: dimension of the key, query, and value.
        out_features: dimension of the last projection
        broadcast_dropout: bool: use a broadcasted dropout along batch dims.
        dropout_rate: dropout rate
        deterministic: if false, the attention weight is masked randomly using
            dropout, whereas if true, the attention weights are deterministic.
        precision: numerical precision of the computation see `jax.lax.Precision`
            for details.
        kernel_init: initializer for the kernel of the Dense layers.
        bias_init: initializer for the bias of the Dense layers.
        use_bias: bool: whether pointwise QKVO dense transforms use bias.
        attention_fn: dot_product_attention or compatible function. Accepts query,
            key, value, and returns output of shape `[bs, dim1, dim2, ..., dimN,,
            num_heads, value_channels]``
        decode: whether to prepare and use an autoregressive cache.
    """

    num_heads: int
    dtype: Optional[Dtype] = None
    param_dtype: Dtype = jnp.float32
    qkv_features: Optional[int] = None
    out_features: Optional[int] = None
    broadcast_dropout: bool = True
    dropout_rate: float = 0.0
    deterministic: Optional[bool] = None
    precision: PrecisionLike = None
    kernel_init: Callable[[PRNGKey, Shape, Dtype], Array] = default_kernel_init
    bias_init: Callable[[PRNGKey, Shape, Dtype], Array] = initializers.zeros_init()
    use_bias: bool = True
    attention_fn: Callable[..., Array] = dot_product_attention
    decode: bool = False
    qkv_dot_general: DotGeneralT = lax.dot_general
    out_dot_general: DotGeneralT = lax.dot_general

    @compact
    def __call__(
        self,
        inputs_q: Array,
        inputs_kv: Array,
        pos_embed: Array,
        mask: Optional[Array] = None,
        deterministic: Optional[bool] = None,
    ):
        """Applies multi-head dot product attention on the input data.

        Projects the inputs into multi-headed query, key, and value vectors,
        applies dot-product attention and project the results to an output vector.

        Args:
            inputs_q: input queries of shape
                `[batch_sizes..., length, features]`.
            inputs_kv: key/values of shape
                `[batch_sizes..., length, features]`.
            mask: attention mask of shape
                `[batch_sizes..., num_heads, query_length, key/value_length]`.
                Attention weights are masked out if their corresponding mask value
                is `False`.
            deterministic: if false, the attention weight is masked randomly
                using dropout, whereas if true, the attention weights
                are deterministic.

        Returns:
            output of shape `[batch_sizes..., length, features]`.
        """
        features = self.out_features or inputs_q.shape[-1]
        qkv_features = self.qkv_features or inputs_q.shape[-1]
        assert qkv_features % self.num_heads == 0, (
            f"Memory dimension ({qkv_features}) must be divisible by number of" f" heads ({self.num_heads})."
        )
        head_dim = qkv_features // self.num_heads

        dense = functools.partial(
            DenseGeneral,
            axis=-1,
            dtype=self.dtype,
            param_dtype=self.param_dtype,
            features=(self.num_heads, head_dim),
            kernel_init=self.kernel_init,
            bias_init=self.bias_init,
            use_bias=self.use_bias,
            precision=self.precision,
            dot_general=self.qkv_dot_general,
        )
        # project inputs_q to multi-headed q/k/v
        # dimensions are then [batch..., length, n_heads, n_features_per_head]
        query, key, value = (
            dense(name="query")(inputs_q),
            dense(name="key")(inputs_kv),
            dense(name="value")(inputs_kv),
        )

        # different bc no bias
        dense_relpos = functools.partial(
            DenseGeneral,
            axis=-1,
            dtype=self.dtype,
            param_dtype=self.param_dtype,
            features=(self.num_heads, head_dim),
            kernel_init=self.kernel_init,
            use_bias=False,
            precision=self.precision,
            dot_general=self.qkv_dot_general,
        )

        r_pos_embed = dense_relpos(name="pos_embed_mat")(pos_embed)

        r_r_bias = self.param("r_r_bias", self.bias_init, (self.num_heads, head_dim))  # Initialization function
        r_w_bias = self.param("r_w_bias", self.bias_init, (self.num_heads, head_dim))  # Initialization function

        # During fast autoregressive decoding, we feed one position at a time,
        # and cache the keys and values step by step.
        if self.decode:
            # detect if we're initializing by absence of existing cache data.
            is_initialized = self.has_variable("cache", "cached_key")
            cached_key = self.variable("cache", "cached_key", jnp.zeros, key.shape, key.dtype)
            cached_value = self.variable("cache", "cached_value", jnp.zeros, value.shape, value.dtype)
            cache_index = self.variable("cache", "cache_index", lambda: jnp.array(0, dtype=jnp.int32))
            if is_initialized:
                (
                    *batch_dims,
                    max_length,
                    num_heads,
                    depth_per_head,
                ) = cached_key.value.shape
                # shape check of cached keys against query input
                expected_shape = tuple(batch_dims) + (1, num_heads, depth_per_head)
                if expected_shape != query.shape:
                    raise ValueError(
                        "Autoregressive cache shape error, "
                        "expected query shape %s instead got %s." % (expected_shape, query.shape)
                    )
                # update key, value caches with our new 1d spatial slices
                cur_index = cache_index.value
                indices = (0,) * len(batch_dims) + (cur_index, 0, 0)
                key = lax.dynamic_update_slice(cached_key.value, key, indices)
                value = lax.dynamic_update_slice(cached_value.value, value, indices)
                cached_key.value = key
                cached_value.value = value
                cache_index.value = cache_index.value + 1
                # causal mask for cached decoder self-attention:
                # our single query position should only attend to those key
                # positions that have already been generated and cached,
                # not the remaining zero elements.
                mask = combine_masks(
                    mask,
                    jnp.broadcast_to(
                        jnp.arange(max_length) <= cur_index,
                        tuple(batch_dims) + (1, 1, max_length),
                    ),
                )

        dropout_rng = None
        if self.dropout_rate > 0.0:  # Require `deterministic` only if using dropout.
            m_deterministic = merge_param("deterministic", self.deterministic, deterministic)
            if not m_deterministic:
                dropout_rng = self.make_rng("dropout")
        else:
            m_deterministic = True

        # apply attention
        x = self.attention_fn(
            query,
            key,
            value,
            r_pos_embed,
            r_r_bias,
            r_w_bias,
            mask=mask,
            dropout_rng=dropout_rng,
            dropout_rate=self.dropout_rate,
            broadcast_dropout=self.broadcast_dropout,
            deterministic=m_deterministic,
            dtype=self.dtype,
            precision=self.precision,
        )  # pytype: disable=wrong-keyword-args
        # back to the original inputs dimensions
        out = DenseGeneral(
            features=features,
            axis=(-2, -1),
            kernel_init=self.kernel_init,
            bias_init=self.bias_init,
            use_bias=self.use_bias,
            dtype=self.dtype,
            param_dtype=self.param_dtype,
            precision=self.precision,
            dot_general=self.out_dot_general,
            name="out",  # type: ignore[call-arg]
        )(x)
        return out


class SelfAttention(RelMultiHeadDotProductAttention):
    """Self-attention special case of multi-head dot-product attention."""

    @compact
    def __call__(  # type: ignore
        self,
        inputs_q: Array,
        mask: Optional[Array] = None,
        deterministic: Optional[bool] = None,
    ):
        """Applies multi-head dot product self-attention on the input data.

        Projects the inputs into multi-headed query, key, and value vectors,
        applies dot-product attention and project the results to an output vector.

        Args:
            inputs_q: input queries of shape
                `[batch_sizes..., length, features]`.
            mask: attention mask of shape
                `[batch_sizes..., num_heads, query_length, key/value_length]`.
                Attention weights are masked out if their corresponding mask value
                is `False`.
            deterministic: if false, the attention weight is masked randomly
                using dropout, whereas if true, the attention weights
                are deterministic.

        Returns:
            output of shape `[batch_sizes..., length, features]`.
        """
        return super().__call__(inputs_q, inputs_q, mask, deterministic=deterministic)


# mask-making utility functions


def make_attention_mask(
    query_input: Array,
    key_input: Array,
    pairwise_fn: Callable[..., Any] = jnp.multiply,
    extra_batch_dims: int = 0,
    dtype: Dtype = jnp.float32,
):
    """Mask-making helper for attention weights.

    In case of 1d inputs (i.e., `[batch..., len_q]`, `[batch..., len_kv]`, the
    attention weights will be `[batch..., heads, len_q, len_kv]` and this
    function will produce `[batch..., 1, len_q, len_kv]`.

    Args:
        query_input: a batched, flat input of query_length size
        key_input: a batched, flat input of key_length size
        pairwise_fn: broadcasting elementwise comparison function
        extra_batch_dims: number of extra batch dims to add singleton
            axes for, none by default
        dtype: mask return dtype

    Returns:
        A `[batch..., 1, len_q, len_kv]` shaped mask for 1d attention.
    """
    mask = pairwise_fn(jnp.expand_dims(query_input, axis=-1), jnp.expand_dims(key_input, axis=-2))
    mask = jnp.expand_dims(mask, axis=-3)
    mask = jnp.expand_dims(mask, axis=tuple(range(extra_batch_dims)))
    return mask.astype(dtype)


def make_causal_mask(x: Array, extra_batch_dims: int = 0, dtype: Dtype = jnp.float32) -> Array:
    """Make a causal mask for self-attention.

    In case of 1d inputs (i.e., `[batch..., len]`, the self-attention weights
    will be `[batch..., heads, len, len]` and this function will produce a
    causal mask of shape `[batch..., 1, len, len]`.

    Args:
        x: input array of shape `[batch..., len]`
        extra_batch_dims: number of batch dims to add singleton axes for,
            none by default
        dtype: mask return dtype

    Returns:
        A `[batch..., 1, len, len]` shaped causal mask for 1d attention.
    """
    idxs = jnp.broadcast_to(jnp.arange(x.shape[-1], dtype=jnp.int32), x.shape)
    return make_attention_mask(
        idxs,
        idxs,
        jnp.greater_equal,
        extra_batch_dims=extra_batch_dims,
        dtype=dtype,
    )


def combine_masks(*masks: Optional[Array], dtype: Dtype = jnp.float32) -> Array:
    """Combine attention masks.

    Args:
        *masks: set of attention mask arguments to combine, some can be None.
        dtype: dtype for the returned mask.

    Returns:
        Combined mask, reduced by logical and, returns None if no masks given.
    """
    masks_list = [m for m in masks if m is not None]
    if not masks_list:
        return None
    assert all(
        map(lambda x: x.ndim == masks_list[0].ndim, masks_list)
    ), f"masks must have same rank: {tuple(map(lambda x: x.ndim, masks_list))}"
    mask, *other_masks = masks_list
    for other_mask in other_masks:
        mask = jnp.logical_and(mask, other_mask)
    return mask.astype(dtype)



from typing import Any, Sequence

import distrax
import jax
import jax.numpy as jnp
from chex import PRNGKey
from flax import struct


@struct.dataclass
class HybridAction:
    discrete: int
    continuous: jnp.ndarray


class HybridActionDistribution(distrax.Distribution):
    def __init__(self, discrete_logits, continuous_mu, continuous_sigma) -> None:
        self.discrete = distrax.Categorical(logits=discrete_logits)
        self.continuous = distrax.MultivariateNormalDiag(continuous_mu, continuous_sigma)

    def _sample_n(self, rng: PRNGKey, n: int) -> Any:
        rng, _rng, _rng2 = jax.random.split(rng, 3)
        a = self.discrete._sample_n(_rng, n)
        b = self.continuous._sample_n(_rng2, n)
        return HybridAction(a, b)

    def log_prob(self, value: Any):
        a = self.discrete.log_prob(value.discrete)
        b = self.continuous.log_prob(value.continuous)
        return a + b  # log probs, we add.

    def entropy(self):
        return self.discrete.entropy() + self.continuous.entropy()

    def event_shape(self) -> Sequence[int]:
        return ()


class MultiDiscreteActionDistribution(distrax.Distribution):
    def __init__(self, flat_logits, number_of_dims_per_distribution) -> None:
        self.distributions = []
        total_dims = 0
        for dims in number_of_dims_per_distribution:
            self.distributions.append(distrax.Categorical(logits=flat_logits[..., total_dims : total_dims + dims]))
            total_dims += dims

    def _sample_n(self, key: PRNGKey, n: int) -> Any:
        rngs = jax.random.split(key, len(self.distributions))
        samples = [jnp.expand_dims(d._sample_n(rng, n), axis=-1) for rng, d in zip(rngs, self.distributions)]
        return jnp.concatenate(samples, axis=-1)

    def log_prob(self, value: Any):
        return sum(d.log_prob(value[..., i]) for i, d in enumerate(self.distributions))

    def entropy(self):
        return sum(d.entropy() for d in self.distributions)

    def event_shape(self) -> Sequence[int]:
        return ()



from timeit import default_timer as tmr

counter = 0


def time_function(f, name):
    global counter
    t = "\t" * counter
    # print(f"{t}Starting... {name}")
    ss = tmr()
    counter += 1
    a = f()
    counter -= 1
    print(f"{t}{name} took {tmr() - ss} seconds")
    return a



import os
from posixpath import isabs
from functools import partial
from typing import Tuple
import typing

import chex
import jax
import jax.numpy as jnp
from flax.training.train_state import TrainState

from gymnax import EnvParams, EnvState
from gymnax.environments.environment import Environment
from kinetix.models.actor_critic import ScannedRNN

Observation = typing.Any


def evaluate_rnn(  # from jaxued
    rng: chex.PRNGKey,
    env: Environment,
    env_params: EnvParams,
    train_state: TrainState,
    init_hstate: chex.ArrayTree,
    init_obs: Observation,
    init_env_state: EnvState,
    max_episode_length: int,
    keep_states=True,
    return_trajectories=False,
) -> Tuple[chex.Array, chex.Array, chex.Array]:
    """This runs the RNN on the environment, given an initial state and observation, and returns (states, rewards, episode_lengths)

    Args:
        rng (chex.PRNGKey):
        env (Environment):
        env_params (EnvParams):
        train_state (TrainState):
        init_hstate (chex.ArrayTree): Shape (num_levels, )
        init_obs (Observation): Shape (num_levels, )
        init_env_state (EnvState): Shape (num_levels, )
        max_episode_length (int):

    Returns:
        Tuple[chex.Array, chex.Array, chex.Array]: (States, rewards, episode lengths) ((NUM_STEPS, NUM_LEVELS), (NUM_STEPS, NUM_LEVELS), (NUM_LEVELS,)
    """
    num_levels = jax.tree_util.tree_flatten(init_obs)[0][0].shape[0]

    def step(carry, _):
        rng, hstate, obs, state, done, mask, episode_length = carry
        rng, rng_action, rng_step = jax.random.split(rng, 3)

        x = jax.tree.map(lambda x: x[None, ...], (obs, done))
        hstate, pi, _ = train_state.apply_fn(train_state.params, hstate, x)
        action = pi.sample(seed=rng_action).squeeze(0)

        obs, next_state, reward, done, info = jax.vmap(env.step, in_axes=(0, 0, 0, None, 0))(
            jax.random.split(rng_step, num_levels),
            state,
            action,
            env_params,
            init_env_state,  # use this to reset to the init env state, keyword arguments are not easily vmappable
        )

        next_mask = mask & ~done
        episode_length += mask

        if keep_states:
            return (rng, hstate, obs, next_state, done, next_mask, episode_length), (state, reward, done, info)
        else:
            return (rng, hstate, obs, next_state, done, next_mask, episode_length), (None, reward, done, info)

    (_, _, _, _, _, _, episode_lengths), (states, rewards, dones, infos) = jax.lax.scan(
        step,
        (
            rng,
            init_hstate,
            init_obs,
            init_env_state,
            jnp.zeros(num_levels, dtype=bool),
            jnp.ones(num_levels, dtype=bool),
            jnp.zeros(num_levels, dtype=jnp.int32),
        ),
        None,
        length=max_episode_length,
    )
    done_idx = jnp.argmax(dones, axis=0)

    to_return = (states, rewards, done_idx, episode_lengths, infos)
    if return_trajectories:
        return to_return, (dones, rewards)
    return to_return


def general_eval(
    rng: chex.PRNGKey,
    eval_env: Environment,
    env_params: EnvParams,
    train_state: TrainState,
    levels: EnvState,
    num_eval_steps: int,
    num_levels: int,
    keep_states=True,
    return_trajectories=False,
    initialize_carry_fn=ScannedRNN.initialize_carry,
):
    """
    This evaluates the current policy on the set of evaluation levels
    It returns (states, cum_rewards, episode_lengths), with shapes (num_steps, num_eval_levels, ...), (num_eval_levels,), (num_eval_levels,)
    """
    rng, rng_reset = jax.random.split(rng)
    init_obs, init_env_state = jax.vmap(eval_env.reset, (0, None, 0))(
        jax.random.split(rng_reset, num_levels), env_params, levels
    )
    init_hstate = initialize_carry_fn(num_levels)
    (states, rewards, done_idx, episode_lengths, infos), (dones, reward) = evaluate_rnn(
        rng,
        eval_env,
        env_params,
        train_state,
        init_hstate,
        init_obs,
        init_env_state,
        num_eval_steps,
        keep_states=keep_states,
        return_trajectories=True,
    )
    mask = jnp.arange(num_eval_steps)[..., None] < episode_lengths
    cum_rewards = (rewards * mask).sum(axis=0)
    to_return = (
        states,
        cum_rewards,
        done_idx,
        episode_lengths,
        infos,
    )  # (num_steps, num_eval_levels, ...), (num_eval_levels,), (num_eval_levels,)

    if return_trajectories:
        return to_return, (dones, reward)
    return to_return


def compute_gae(
    gamma: float,
    lambd: float,
    last_value: chex.Array,
    values: chex.Array,
    rewards: chex.Array,
    dones: chex.Array,
) -> Tuple[chex.Array, chex.Array]:
    """This takes in arrays of shape (NUM_STEPS, NUM_ENVS) and returns the advantages and targets.

    Args:
        gamma (float):
        lambd (float):
        last_value (chex.Array):  Shape (NUM_ENVS)
        values (chex.Array): Shape (NUM_STEPS, NUM_ENVS)
        rewards (chex.Array): Shape (NUM_STEPS, NUM_ENVS)
        dones (chex.Array): Shape (NUM_STEPS, NUM_ENVS)

    Returns:
        Tuple[chex.Array, chex.Array]: advantages, targets; each of shape (NUM_STEPS, NUM_ENVS)
    """

    def compute_gae_at_timestep(carry, x):
        gae, next_value = carry
        value, reward, done = x
        delta = reward + gamma * next_value * (1 - done) - value
        gae = delta + gamma * lambd * (1 - done) * gae
        return (gae, value), gae

    _, advantages = jax.lax.scan(
        compute_gae_at_timestep,
        (jnp.zeros_like(last_value), last_value),
        (values, rewards, dones),
        reverse=True,
        unroll=16,
    )
    return advantages, advantages + values


def sample_trajectories_rnn(
    rng: chex.PRNGKey,
    env: Environment,
    env_params: EnvParams,
    train_state: TrainState,
    init_hstate: chex.ArrayTree,
    init_obs: Observation,
    init_env_state: EnvState,
    num_envs: int,
    max_episode_length: int,
    return_states: bool = False,
) -> Tuple[
    Tuple[chex.PRNGKey, TrainState, chex.ArrayTree, Observation, EnvState, chex.Array],
    Tuple[Observation, chex.Array, chex.Array, chex.Array, chex.Array, chex.Array, dict],
]:
    """This samples trajectories from the environment using the agent specified by the `train_state`.

    Args:

        rng (chex.PRNGKey): Singleton
        env (Environment):
        env_params (EnvParams):
        train_state (TrainState): Singleton
        init_hstate (chex.ArrayTree): This is the init RNN hidden state, has to have shape (NUM_ENVS, ...)
        init_obs (Observation): The initial observation, shape (NUM_ENVS, ...)
        init_env_state (EnvState): The initial env state (NUM_ENVS, ...)
        num_envs (int): The number of envs that are vmapped over.
        max_episode_length (int): The maximum episode length, i.e., the number of steps to do the rollouts for.

    Returns:
        Tuple[Tuple[chex.PRNGKey, TrainState, chex.ArrayTree, Observation, EnvState, chex.Array], Tuple[Observation, chex.Array, chex.Array, chex.Array, chex.Array, chex.Array, dict]]: (rng, train_state, hstate, last_obs, last_env_state, last_value), traj, where traj is (obs, action, reward, done, log_prob, value, info). The first element in the tuple consists of arrays that have shapes (NUM_ENVS, ...) (except `rng` and and `train_state` which are singleton). The second element in the tuple is of shape (NUM_STEPS, NUM_ENVS, ...), and it contains the trajectory.
    """

    def sample_step(carry, _):
        rng, train_state, hstate, obs, env_state, last_done = carry
        prev_state = env_state
        rng, rng_action, rng_step = jax.random.split(rng, 3)

        x = jax.tree.map(lambda x: x[None, ...], (obs, last_done))
        hstate, pi, value = train_state.apply_fn(train_state.params, hstate, x)
        action = pi.sample(seed=rng_action)
        log_prob = pi.log_prob(action)
        value, action, log_prob = jax.tree.map(lambda x: x.squeeze(0), (value, action, log_prob))

        next_obs, env_state, reward, done, info = jax.vmap(env.step, in_axes=(0, 0, 0, None, 0))(
            jax.random.split(rng_step, num_envs), env_state, action, env_params, init_env_state
        )

        carry = (rng, train_state, hstate, next_obs, env_state, done)
        step = (obs, action, reward, done, log_prob, value, info)
        if return_states:
            step += (prev_state,)
        return carry, step

    (rng, train_state, hstate, last_obs, last_env_state, last_done), traj = jax.lax.scan(
        sample_step,
        (
            rng,
            train_state,
            init_hstate,
            init_obs,
            init_env_state,
            jnp.zeros(num_envs, dtype=bool),
        ),
        None,
        length=max_episode_length,
    )

    x = jax.tree.map(lambda x: x[None, ...], (last_obs, last_done))
    _, _, last_value = train_state.apply_fn(train_state.params, hstate, x)

    my_obs = traj[0]
    rew = traj[2]

    return (rng, train_state, hstate, last_obs, last_env_state, last_value.squeeze(0)), traj


def update_actor_critic_rnn(
    rng: chex.PRNGKey,
    train_state: TrainState,
    init_hstate: chex.ArrayTree,
    batch: chex.ArrayTree,
    num_envs: int,
    n_steps: int,
    n_minibatch: int,
    n_epochs: int,
    clip_eps: float,
    entropy_coeff: float,
    critic_coeff: float,
    update_grad: bool = True,
) -> Tuple[Tuple[chex.PRNGKey, TrainState], chex.ArrayTree]:
    """This function takes in a rollout, and PPO hyperparameters, and updates the train state.

    Args:
        rng (chex.PRNGKey):
        train_state (TrainState):
        init_hstate (chex.ArrayTree):
        batch (chex.ArrayTree): obs, actions, dones, log_probs, values, targets, advantages
        num_envs (int):
        n_steps (int):
        n_minibatch (int):
        n_epochs (int):
        clip_eps (float):
        entropy_coeff (float):
        critic_coeff (float):
        update_grad (bool, optional): If False, the train state does not actually get updated. Defaults to True.

    Returns:
        Tuple[Tuple[chex.PRNGKey, TrainState], chex.ArrayTree]: It returns a new rng, the updated train_state, and the losses. The losses have structure (loss, (l_vf, l_clip, entropy))
    """
    obs, actions, dones, log_probs, values, targets, advantages = batch
    last_dones = jnp.roll(dones, 1, axis=0).at[0].set(False)
    batch = obs, actions, last_dones, log_probs, values, targets, advantages

    def update_epoch(carry, _):
        def update_minibatch(train_state, minibatch):
            init_hstate, obs, actions, last_dones, log_probs, values, targets, advantages = minibatch

            def loss_fn(params):
                _, pi, values_pred = train_state.apply_fn(params, init_hstate, (obs, last_dones))
                log_probs_pred = pi.log_prob(actions)
                entropy = pi.entropy().mean()

                ratio = jnp.exp(log_probs_pred - log_probs)
                A = (advantages - advantages.mean()) / (advantages.std() + 1e-8)
                l_clip = (-jnp.minimum(ratio * A, jnp.clip(ratio, 1 - clip_eps, 1 + clip_eps) * A)).mean()

                values_pred_clipped = values + (values_pred - values).clip(-clip_eps, clip_eps)
                l_vf = 0.5 * jnp.maximum((values_pred - targets) ** 2, (values_pred_clipped - targets) ** 2).mean()

                loss = l_clip + critic_coeff * l_vf - entropy_coeff * entropy

                return loss, (l_vf, l_clip, entropy)

            grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
            loss, grads = grad_fn(train_state.params)
            if update_grad:
                train_state = train_state.apply_gradients(grads=grads)
            grad_norm = jnp.linalg.norm(
                jnp.concatenate(jax.tree.map(lambda x: x.flatten(), jax.tree_util.tree_flatten(grads)[0]))
            )
            return train_state, (loss, grad_norm)

        rng, train_state = carry
        rng, rng_perm = jax.random.split(rng)
        permutation = jax.random.permutation(rng_perm, num_envs)
        minibatches = (
            jax.tree.map(
                lambda x: jnp.take(x, permutation, axis=0).reshape(n_minibatch, -1, *x.shape[1:]),
                init_hstate,
            ),
            *jax.tree.map(
                lambda x: jnp.take(x, permutation, axis=1)
                .reshape(x.shape[0], n_minibatch, -1, *x.shape[2:])
                .swapaxes(0, 1),
                batch,
            ),
        )
        train_state, (losses, grads) = jax.lax.scan(update_minibatch, train_state, minibatches)
        return (rng, train_state), (losses, grads)

    return jax.lax.scan(update_epoch, (rng, train_state), None, n_epochs)


# Cannot jit outside otherwise jax errors with trying to hash a tracer.
# @partial(jax.jit, static_argnums=(0, 2, 8, 9))
def sample_trajectories_and_learn(
    env: Environment,
    env_params: EnvParams,
    config: dict,
    rng: chex.PRNGKey,
    train_state: TrainState,
    init_hstate: chex.Array,
    init_obs: Observation,
    init_env_state: EnvState,
    update_grad: bool = True,
    return_states: bool = False,
) -> Tuple[
    Tuple[chex.PRNGKey, TrainState, Observation, EnvState],
    Tuple[
        Observation,
        chex.Array,
        chex.Array,
        chex.Array,
        chex.Array,
        chex.Array,
        dict,
        chex.Array,
        chex.Array,
        chex.ArrayTree,
        chex.Array,
    ],
]:
    """This function loops the following:
        - rollout for config['num_steps']
        - learn / update policy

    And it loops it for config['outer_rollout_steps'].
    What is returns is a new carry (rng, train_state, init_obs, init_env_state), and concatenated rollouts. The shape of the rollouts are config['num_steps'] * config['outer_rollout_steps']. In other words, the trajectories returned by this function are the same as if we ran rollouts for config['num_steps'] * config['outer_rollout_steps'] steps, but the agent does perform PPO updates in between.

    Args:
        env (Environment):
        env_params (EnvParams):
        config (dict):
        rng (chex.PRNGKey):
        train_state (TrainState):
        init_obs (Observation):
        init_env_state (EnvState):
        update_grad (bool, optional): Defaults to True.

    Returns:
        Tuple[Tuple[chex.PRNGKey, TrainState, Observation, EnvState], Tuple[Observation, chex.Array, chex.Array, chex.Array, chex.Array, chex.Array, dict, chex.Array, chex.Array, chex.ArrayTree, chex.Array]]: This returns a tuple:
        (
            (rng, train_state, init_obs, init_env_state),
            (obs, actions, rewards, dones, log_probs, values, info, advantages, targets, losses, grads)
        )
    """

    @jax.jit
    def single_step(carry, _):
        rng, train_state, init_hstate, init_obs, init_env_state = carry
        ((rng, train_state, new_hstate, last_obs, last_env_state, last_value), traj,) = sample_trajectories_rnn(
            rng,
            env,
            env_params,
            train_state,
            init_hstate,
            init_obs,
            init_env_state,
            config["num_train_envs"],
            config["num_steps"],
            return_states=return_states,
        )
        if return_states:
            states = traj[-1]
            traj = traj[:-1]

        (obs, actions, rewards, dones, log_probs, values, info) = traj
        advantages, targets = compute_gae(config["gamma"], config["gae_lambda"], last_value, values, rewards, dones)

        # Update the policy using trajectories collected from replay levels
        (rng, train_state), (losses, grads) = update_actor_critic_rnn(
            rng,
            train_state,
            init_hstate,
            (obs, actions, dones, log_probs, values, targets, advantages),
            config["num_train_envs"],
            config["num_steps"],
            config["num_minibatches"],
            config["update_epochs"],
            config["clip_eps"],
            config["ent_coef"],
            config["vf_coef"],
            update_grad=update_grad,
        )
        new_carry = (rng, train_state, new_hstate, last_obs, last_env_state)
        step = (obs, actions, rewards, dones, log_probs, values, info, advantages, targets, losses, grads)
        if return_states:
            step += (states,)
        return new_carry, step

    @jax.jit
    def inner(rng, train_state, init_hstate, init_obs, init_env_state):
        carry = (rng, train_state, init_hstate, init_obs, init_env_state)
        new_carry, all_rollouts = jax.lax.scan(single_step, carry, None, length=config["outer_rollout_steps"])

        all_rollouts = jax.tree.map(lambda x: jnp.concatenate(x, axis=0), all_rollouts)
        return new_carry, all_rollouts

    return inner(rng, train_state, init_hstate, init_obs, init_env_state)


def no_op_rollout(
    env: Environment,
    env_params: EnvParams,
    rng: chex.PRNGKey,
    init_obs: Observation,
    init_env_state: EnvState,
    num_envs: int,
    max_episode_length: int,
    do_random=False,
):

    noop = jnp.array(env.action_type.noop_action())
    zero_action = jnp.repeat(noop[None, ...], num_envs, axis=0)
    SHAPE = zero_action.shape

    def sample_step(carry, _):
        rng, obs, env_state, last_done = carry
        rng, rng_step, _rng = jax.random.split(rng, 3)
        if do_random:
            action = jax.vmap(env.action_space(env_params).sample)(jax.random.split(_rng, num_envs))
        else:
            action = zero_action

        next_obs, env_state, reward, done, info = jax.vmap(env.step, in_axes=(0, 0, 0, None, 0))(
            jax.random.split(rng_step, num_envs), env_state, action, env_params, init_env_state
        )

        carry = (rng, next_obs, env_state, done)
        return carry, (obs, action, reward, done, info)

    (rng, last_obs, last_env_state, last_done), traj = jax.lax.scan(
        sample_step,
        (
            rng,
            init_obs,
            init_env_state,
            jnp.zeros(num_envs, dtype=bool),
        ),
        None,
        length=max_episode_length,
    )

    info = traj[-1]
    dones = traj[-2]

    returns_per_env = (info["returned_episode_returns"] * dones).sum(axis=0) / jnp.maximum(1, dones.sum(axis=0))
    lens_per_env = (info["returned_episode_lengths"] * dones).sum(axis=0) / jnp.maximum(1, dones.sum(axis=0))
    success_per_env = (info["returned_episode_solved"] * dones).sum(axis=0) / jnp.maximum(1, dones.sum(axis=0))
    return returns_per_env, lens_per_env, success_per_env


def no_op_and_random_rollout(
    env: Environment,
    env_params: EnvParams,
    rng: chex.PRNGKey,
    init_obs: Observation,
    init_env_state: EnvState,
    num_envs: int,
    max_episode_length: int,
):
    returns_noop, lens_noop, success_noop = no_op_rollout(
        env, env_params, rng, init_obs, init_env_state, num_envs, max_episode_length, do_random=False
    )
    returns_random, lens_random, success_random = no_op_rollout(
        env, env_params, rng, init_obs, init_env_state, num_envs, max_episode_length, do_random=True
    )
    return returns_noop, lens_noop, success_noop, returns_random, lens_random, success_random



from typing import List, Sequence

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from flax.linen.attention import MultiHeadDotProductAttention
from flax.linen.initializers import constant, orthogonal

from kinetix.models.actor_critic import GeneralActorCriticRNN
from kinetix.render.renderer_symbolic_entity import EntityObservation


class Gating(nn.Module):
    # code taken from https://github.com/dhruvramani/Transformers-RL/blob/master/layers.py
    d_input: int
    bg: float = 0.0

    @nn.compact
    def __call__(self, x, y):
        r = jax.nn.sigmoid(nn.Dense(self.d_input, use_bias=False)(y) + nn.Dense(self.d_input, use_bias=False)(x))
        z = jax.nn.sigmoid(
            nn.Dense(self.d_input, use_bias=False)(y)
            + nn.Dense(self.d_input, use_bias=False)(x)
            - self.param("gating_bias", constant(self.bg), (self.d_input,))
        )
        h = jnp.tanh(nn.Dense(self.d_input, use_bias=False)(y) + nn.Dense(self.d_input, use_bias=False)(r * x))
        g = (1 - z) * x + (z * h)
        return g


class transformer_layer(nn.Module):
    num_heads: int
    out_features: int
    qkv_features: int
    gating: bool = False
    gating_bias: float = 0.0

    def setup(self):
        self.attention1 = MultiHeadDotProductAttention(
            num_heads=self.num_heads, qkv_features=self.qkv_features, out_features=self.out_features
        )

        self.ln1 = nn.LayerNorm()

        self.dense1 = nn.Dense(self.out_features)

        self.dense2 = nn.Dense(self.out_features)

        self.ln2 = nn.LayerNorm()
        if self.gating:
            self.gate1 = Gating(self.out_features, self.gating_bias)
            self.gate2 = Gating(self.out_features, self.gating_bias)

    def __call__(self, queries: jnp.ndarray, mask: jnp.ndarray):
        # After reading the paper, this is what I think we should do:
        # First layernorm, then do attention
        queries_n = self.ln1(queries)
        y = self.attention1(queries_n, mask=mask)
        if self.gating:  # and gate
            y = self.gate1(queries, jax.nn.relu(y))
        else:
            y = queries + y
        # Dense after norming, crucially no relu.
        e = self.dense1(self.ln2(y))
        if self.gating:  # and gate again
            # This may be the wrong way around
            e = self.gate2(y, jax.nn.relu(e))
        else:
            e = y + e

        return e


class Transformer(nn.Module):
    encoder_size: int
    num_heads: int
    qkv_features: int
    num_layers: int
    gating: bool = False
    gating_bias: float = 0.0

    def setup(self):
        # self.encoder = nn.Dense(self.encoder_size)

        # self.positional_encoding = PositionalEncoding(self.encoder_size, max_len=self.max_len)

        self.tf_layers = [
            transformer_layer(
                num_heads=self.num_heads,
                qkv_features=self.qkv_features,
                out_features=self.encoder_size,
                gating=self.gating,
                gating_bias=self.gating_bias,
            )
            for _ in range(self.num_layers)
        ]

        self.joint_layers = [nn.Dense(self.encoder_size) for _ in range(self.num_layers)]
        self.thruster_layers = [nn.Dense(self.encoder_size) for _ in range(self.num_layers)]

        # self.pos_emb=PositionalEmbedding(self.encoder_size)

    def __call__(
        self,
        shape_embeddings: jnp.ndarray,
        shape_attention_mask,
        joint_embeddings,
        joint_mask,
        joint_indexes,
        thruster_embeddings,
        thruster_mask,
        thruster_indexes,
    ):
        # forward eval so obs is only one timestep
        # encoded = self.encoder(shape_embeddings)
        # pos_embed=self.pos_emb(jnp.arange(1+memories.shape[-3],-1,-1))[:1+memories.shape[-3]]

        for tf_layer, joint_layer, thruster_layer in zip(self.tf_layers, self.joint_layers, self.thruster_layers):
            # Do attention
            shape_embeddings = tf_layer(shape_embeddings, shape_attention_mask)

            # Joints
            # T, B, 2J, (2SE + JE)

            @jax.vmap
            @jax.vmap
            def do_index2(to_ind, ind):
                return to_ind[ind]

            joint_shape_embeddings = jnp.concatenate(
                [
                    do_index2(shape_embeddings, joint_indexes[..., 0]),
                    do_index2(shape_embeddings, joint_indexes[..., 1]),
                    joint_embeddings,
                ],
                axis=-1,
            )

            shape_joint_entity_delta = joint_layer(joint_shape_embeddings) * joint_mask[..., None]

            @jax.vmap
            @jax.vmap
            def add2(addee, index, adder):
                return addee.at[index].add(adder)

            # Thrusters
            thruster_shape_embeddings = jnp.concatenate(
                [
                    do_index2(shape_embeddings, thruster_indexes),
                    thruster_embeddings,
                ],
                axis=-1,
            )

            shape_thruster_entity_delta = thruster_layer(thruster_shape_embeddings) * thruster_mask[..., None]

            shape_embeddings = add2(shape_embeddings, joint_indexes[..., 0], shape_joint_entity_delta)
            shape_embeddings = add2(shape_embeddings, thruster_indexes, shape_thruster_entity_delta)

        return shape_embeddings


class ActorCriticTransformer(nn.Module):
    action_dim: Sequence[int]
    fc_layer_width: int
    action_mode: str
    hybrid_action_continuous_dim: int
    multi_discrete_number_of_dims_per_distribution: List[int]
    transformer_size: int
    transformer_encoder_size: int
    transformer_depth: int
    fc_layer_depth: int
    num_heads: int
    activation: str
    aggregate_mode: str  # "dummy" or "mean" or "dummy_and_mean"
    full_attention_mask: bool  # if true, only mask out inactives, and have everything attend to everything else
    add_generator_embedding: bool = False
    generator_embedding_number_of_timesteps: int = 10
    recurrent: bool = True

    @nn.compact
    def __call__(self, hidden, x):
        if self.activation == "relu":
            activation = nn.relu
        else:
            activation = nn.tanh

        og_obs, dones = x
        if self.add_generator_embedding:
            obs = og_obs.obs
        else:
            obs = og_obs

        # obs._ is [T, B, N, L]
        # B - batch size
        # T - time
        # N - number of things
        # L - unembedded entity size
        obs: EntityObservation

        def _single_encoder(features, entity_id, concat=True):
            # assume two entity types
            num_to_remove = 1 if concat else 0
            embedding = activation(
                nn.Dense(
                    self.transformer_encoder_size - num_to_remove,
                    kernel_init=orthogonal(np.sqrt(2)),
                    bias_init=constant(0.0),
                )(features)
            )
            if concat:
                id_1h = jnp.zeros((*embedding.shape[:3], 1)).at[:, :, :, 0].set(entity_id)
                return jnp.concatenate([embedding, id_1h], axis=-1)
            else:
                return embedding

        circle_encodings = _single_encoder(obs.circles, 0)
        polygon_encodings = _single_encoder(obs.polygons, 1)
        joint_encodings = _single_encoder(obs.joints, -1, False)
        thruster_encodings = _single_encoder(obs.thrusters, -1, False)
        # Size of this is something like (T, B, N, K) (time, batch, num_entities, embedding_size)

        # T, B, M, K
        shape_encodings = jnp.concatenate([polygon_encodings, circle_encodings], axis=2)
        # T, B, M
        shape_mask = jnp.concatenate([obs.polygon_mask, obs.circle_mask], axis=2)

        def mask_out_inactives(flat_active_mask, matrix_attention_mask):
            matrix_attention_mask = matrix_attention_mask & (flat_active_mask[:, None]) & (flat_active_mask[None, :])
            return matrix_attention_mask

        joint_indexes = obs.joint_indexes
        thruster_indexes = obs.thruster_indexes

        if self.aggregate_mode == "dummy" or self.aggregate_mode == "dummy_and_mean":
            T, B, _, K = circle_encodings.shape
            dummy = jnp.ones((T, B, 1, K))
            shape_encodings = jnp.concatenate([dummy, shape_encodings], axis=2)
            shape_mask = jnp.concatenate(
                [jnp.ones((T, B, 1), dtype=bool), shape_mask],
                axis=2,
            )
            N = obs.attention_mask.shape[-1]
            overall_mask = (
                jnp.ones((T, B, obs.attention_mask.shape[2], N + 1, N + 1), dtype=bool)
                .at[:, :, :, 1:, 1:]
                .set(obs.attention_mask)
            )
            overall_mask = jax.vmap(jax.vmap(mask_out_inactives))(shape_mask, overall_mask)

            # To account for the dummy entity
            joint_indexes = joint_indexes + 1
            thruster_indexes = thruster_indexes + 1

        else:
            overall_mask = obs.attention_mask

        if self.full_attention_mask:
            overall_mask = jnp.ones(overall_mask.shape, dtype=bool)
            overall_mask = jax.vmap(jax.vmap(mask_out_inactives))(shape_mask, overall_mask)

        # Now do attention on these
        embedding = Transformer(
            num_layers=self.transformer_depth,
            num_heads=self.num_heads,
            qkv_features=self.transformer_size,
            encoder_size=self.transformer_encoder_size,
            gating=True,
            gating_bias=0.0,
        )(
            shape_encodings,
            jnp.repeat(overall_mask, repeats=self.num_heads // overall_mask.shape[2], axis=2),
            joint_encodings,
            obs.joint_mask,
            joint_indexes,
            thruster_encodings,
            obs.thruster_mask,
            thruster_indexes,
        )  # add the extra dimension for the heads

        if self.aggregate_mode == "mean" or self.aggregate_mode == "dummy_and_mean":
            embedding = jnp.mean(embedding, axis=2, where=shape_mask[..., None])
        else:
            embedding = embedding[:, :, 0]  # Take the dummy entity as the embedding of the entire scene.

        return GeneralActorCriticRNN(
            action_dim=self.action_dim,
            fc_layer_depth=self.fc_layer_depth,
            fc_layer_width=self.fc_layer_width,
            action_type=self.action_mode,
            hybrid_action_continuous_dim=self.hybrid_action_continuous_dim,
            multi_discrete_number_of_dims_per_distribution=self.multi_discrete_number_of_dims_per_distribution,
            add_generator_embedding=self.add_generator_embedding,
            generator_embedding_number_of_timesteps=self.generator_embedding_number_of_timesteps,
            recurrent=self.recurrent,
        )(hidden, og_obs, embedding, dones, activation)
